import argparse
import sys
import subprocess
import re
import os

DEFAULT_UNITTEST_PATH = 'build/release/test/unittest'

parser = argparse.ArgumentParser(description='Print a list of tests to run.')
parser.add_argument(
    '--file-contains',
    dest='file_contains',
    action='store',
    help='Filter based on a string contained in the text',
    default=None,
)
parser.add_argument(
    '--unittest',
    dest='unittest',
    action='store',
    help='The path to the unittest program',
    default=DEFAULT_UNITTEST_PATH,
)
parser.add_argument('--list', dest='filter', action='store', help='The unittest filter to apply', default='')

args = parser.parse_args()

file_contains = args.file_contains
extra_args = [args.filter]
unittest_program = args.unittest

# Override default for windows
if os.name == 'nt' and unittest_program == DEFAULT_UNITTEST_PATH:
    unittest_program = 'build/release/test/Release/unittest.exe'

proc = subprocess.Popen([unittest_program, '-l'] + extra_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
stdout = proc.stdout.read().decode('utf8')
stderr = proc.stderr.read().decode('utf8')
if proc.returncode is not None and proc.returncode != 0:
    print("Failed to run program " + unittest_program)
    print(proc.returncode)
    print(stdout)
    print(stderr)
    exit(1)

test_cases = []
for line in stdout.splitlines()[1:]:
    if not line.strip():
        continue
    splits = line.rsplit('\t', 1)
    if file_contains is not None:
        if not os.path.isfile(splits[0]):
            continue
        try:
            with open(splits[0], 'r') as f:
                text = f.read()
        except UnicodeDecodeError:
            continue
        if file_contains not in text:
            continue
    print(splits[0])



import os
import sys
import subprocess
import threading

if len(sys.argv) < 3:
    print("Expected python3 scripts/try_timeout.py --timeout=[timeout] --retry=[retries] [cmd] [options...]")
    print("Timeout should be given in seconds")
    exit(1)

timeout = int(sys.argv[1].replace("--timeout=", ""))
retries = int(sys.argv[2].replace("--retry=", ""))
cmd = sys.argv[3:]


class Command(object):
    def __init__(self, cmd):
        self.cmd = cmd
        self.process = None

    def run(self, timeout):
        self.process = None

        def target():
            self.process = subprocess.Popen(self.cmd)
            self.process.communicate()

        thread = threading.Thread(target=target)
        thread.start()

        thread.join(timeout)
        if thread.is_alive():
            print('Terminating process: process exceeded timeout of ' + str(timeout) + ' seconds')
            self.process.terminate()
            thread.join()
        if self.process is None:
            return 1
        return self.process.returncode


for i in range(retries):
    print("Attempting to run command \"" + ' '.join(cmd) + '"')
    command = Command(cmd)
    returncode = command.run(timeout)
    if returncode == 0:
        exit(0)

exit(1)



#!/usr/bin/python

# this script is used to format the source directory

import os
import time
import sys
import inspect
import subprocess
import difflib
import re
import tempfile
import uuid
import concurrent.futures
import argparse
import shutil
import traceback
from python_helpers import open_utf8

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from format_test_benchmark import format_file_content

try:
    ver = subprocess.check_output(('black', '--version'), text=True)
    if int(ver.split(' ')[1].split('.')[0]) < 24:
        print('you need to run `pip install "black>=24"`', ver)
        if 'DUCKDB_FORMAT_SKIP_VERSION_CHECKS' not in os.environ:
            exit(-1)
except Exception as e:
    print('you need to run `pip install "black>=24"`', e)
    exit(-1)

try:
    ver = subprocess.check_output(('clang-format', '--version'), text=True)
    if '11.' not in ver:
        print('you need to run `pip install clang_format==11.0.1 - `', ver)
        if 'DUCKDB_FORMAT_SKIP_VERSION_CHECKS' not in os.environ:
            exit(-1)
except Exception as e:
    print('you need to run `pip install clang_format==11.0.1 - `', e)
    exit(-1)

cpp_format_command = 'clang-format --sort-includes=0 -style=file'
cmake_format_command = 'cmake-format'

try:
    subprocess.check_output(('cmake-format', '--version'), text=True)
except Exception as e:
    print('you need to run `pip install cmake-format`', e)
    exit(-1)

extensions = [
    '.cpp',
    '.ipp',
    '.c',
    '.hpp',
    '.h',
    '.cc',
    '.hh',
    'CMakeLists.txt',
    '.test',
    '.test_slow',
    '.test_coverage',
    '.benchmark',
    '.py',
    '.java',
]
formatted_directories = ['src', 'benchmark', 'test', 'tools', 'examples', 'extension', 'scripts']
ignored_files = [
    'tpch_constants.hpp',
    'tpcds_constants.hpp',
    '_generated',
    'test_csv_header.hpp',
    'duckdb.cpp',
    'duckdb.hpp',
    'json.hpp',
    'sqlite3.h',
    'shell.c',
    'termcolor.hpp',
    'test_insert_invalid.test',
    'httplib.hpp',
    'os_win.c',
    'glob.c',
    'printf.c',
    'helper.hpp',
    'single_thread_ptr.hpp',
    'types.hpp',
    'default_views.cpp',
    'default_functions.cpp',
    'release.h',
    'genrand.cpp',
    'address.cpp',
    'visualizer_constants.hpp',
    'icu-collate.cpp',
    'icu-collate.hpp',
    'yyjson.cpp',
    'yyjson.hpp',
    'duckdb_pdqsort.hpp',
    'pdqsort.h',
    'stubdata.cpp',
    'nf_calendar.cpp',
    'nf_calendar.h',
    'nf_localedata.cpp',
    'nf_localedata.h',
    'nf_zformat.cpp',
    'nf_zformat.h',
    'expr.cc',
    'function_list.cpp',
    'inlined_grammar.hpp',
]
ignored_directories = [
    '.eggs',
    '__pycache__',
    'dbgen',
    os.path.join('tools', 'rpkg', 'src', 'duckdb'),
    os.path.join('tools', 'rpkg', 'inst', 'include', 'cpp11'),
    os.path.join('extension', 'tpcds', 'dsdgen'),
    os.path.join('extension', 'icu', 'third_party'),
    os.path.join('tools', 'nodejs', 'src', 'duckdb'),
]
format_all = False
check_only = True
confirm = True
silent = False
force = False


parser = argparse.ArgumentParser(prog='python scripts/format.py', description='Format source directory files')
parser.add_argument(
    'revision', nargs='?', default='HEAD', help='Revision number or --all to format all files (default: HEAD)'
)
parser.add_argument('--check', action='store_true', help='Only print differences (default)')
parser.add_argument('--fix', action='store_true', help='Fix the files')
parser.add_argument('-a', '--all', action='store_true', help='Format all files')
parser.add_argument('-d', '--directories', nargs='*', default=[], help='Format specified directories')
parser.add_argument('-y', '--noconfirm', action='store_true', help='Skip confirmation prompt')
parser.add_argument('-q', '--silent', action='store_true', help='Suppress output')
parser.add_argument('-f', '--force', action='store_true', help='Force formatting')
args = parser.parse_args()

revision = args.revision
if args.check and args.fix:
    parser.print_usage()
    exit(1)
check_only = not args.fix
confirm = not args.noconfirm
silent = args.silent
force = args.force
format_all = args.all
if args.directories:
    formatted_directories = args.directories


def file_is_ignored(full_path):
    if os.path.basename(full_path) in ignored_files:
        return True
    dirnames = os.path.sep.join(full_path.split(os.path.sep)[:-1])
    for ignored_directory in ignored_directories:
        if ignored_directory in dirnames:
            return True
    return False


def can_format_file(full_path):
    global extensions, formatted_directories, ignored_files
    if not os.path.isfile(full_path):
        return False
    fname = full_path.split(os.path.sep)[-1]
    found = False
    # check file extension
    for ext in extensions:
        if full_path.endswith(ext):
            found = True
            break
    if not found:
        return False
    # check ignored files
    if file_is_ignored(full_path):
        return False
    # now check file directory
    for dname in formatted_directories:
        if full_path.startswith(dname):
            return True
    return False


action = "Formatting"
if check_only:
    action = "Checking"


def get_changed_files(revision):
    proc = subprocess.Popen(['git', 'diff', '--name-only', revision], stdout=subprocess.PIPE)
    files = proc.stdout.read().decode('utf8').split('\n')
    changed_files = []
    for f in files:
        if not can_format_file(f):
            continue
        if file_is_ignored(f):
            continue
        changed_files.append(f)
    return changed_files


if os.path.isfile(revision):
    print(action + " individual file: " + revision)
    changed_files = [revision]
elif os.path.isdir(revision):
    print(action + " files in directory: " + revision)
    changed_files = [os.path.join(revision, x) for x in os.listdir(revision)]

    print("Changeset:")
    for fname in changed_files:
        print(fname)
elif not format_all:
    if revision == 'main':
        # fetch new changes when comparing to the master
        os.system("git fetch origin main:main")
    print(action + " since branch or revision: " + revision)
    changed_files = get_changed_files(revision)
    if len(changed_files) == 0:
        print("No changed files found!")
        exit(0)

    print("Changeset:")
    for fname in changed_files:
        print(fname)
else:
    print(action + " all files")

if confirm and not check_only:
    print("The files listed above will be reformatted.")
    result = input("Continue with changes (y/n)?\n")
    if result != 'y':
        print("Aborting.")
        exit(0)

format_commands = {
    '.cpp': cpp_format_command,
    '.ipp': cpp_format_command,
    '.c': cpp_format_command,
    '.hpp': cpp_format_command,
    '.h': cpp_format_command,
    '.hh': cpp_format_command,
    '.cc': cpp_format_command,
    '.txt': cmake_format_command,
    '.py': 'black --quiet - --skip-string-normalization --line-length 120 --stdin-filename',
    '.java': cpp_format_command,
}

difference_files = []

header_top = "//===----------------------------------------------------------------------===//\n"
header_top += "//                         DuckDB\n" + "//\n"
header_bottom = "//\n" + "//\n"
header_bottom += "//===----------------------------------------------------------------------===//\n\n"
base_dir = os.path.join(os.getcwd(), 'src/include')


def get_formatted_text(f, full_path, directory, ext):
    if not can_format_file(full_path):
        if not force:
            print(
                "File "
                + full_path
                + " is not normally formatted - but attempted to format anyway. Use --force if formatting is desirable"
            )
            exit(1)
    if f == 'list.hpp':
        # fill in list file
        file_list = [
            os.path.join(dp, f)
            for dp, dn, filenames in os.walk(directory)
            for f in filenames
            if os.path.splitext(f)[1] == '.hpp' and not f.endswith("list.hpp")
        ]
        file_list = [x.replace('src/include/', '') for x in file_list]
        file_list.sort()
        result = ""
        for x in file_list:
            result += '#include "%s"\n' % (x)
        return result

    if ext == ".hpp" and directory.startswith("src/include"):
        with open_utf8(full_path, 'r') as f:
            lines = f.readlines()

        # format header in files
        header_middle = "// " + os.path.relpath(full_path, base_dir) + "\n"
        text = header_top + header_middle + header_bottom
        is_old_header = True
        for line in lines:
            if not (line.startswith("//") or line.startswith("\n")) and is_old_header:
                is_old_header = False
            if not is_old_header:
                text += line

    if ext == '.test' or ext == '.test_slow' or ext == '.test_coverage' or ext == '.benchmark':
        # optimization: import and call the function directly
        # instead of running a subprocess
        with open(full_path, "r", encoding="utf-8") as f:
            original_lines = f.readlines()
        formatted, status = format_file_content(full_path, original_lines)
        if formatted is None:
            print(f"Failed to format {full_path}: {status}")
            sys.exit(1)
        return formatted
    proc_command = format_commands[ext].split(' ') + [full_path]
    proc = subprocess.Popen(
        proc_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=open(full_path) if ext == '.py' else None
    )
    new_text = proc.stdout.read().decode('utf8')
    stderr = proc.stderr.read().decode('utf8')
    if len(stderr) > 0:
        print(os.getcwd())
        print("Failed to format file " + full_path)
        print(' '.join(proc_command))
        print(stderr)
        exit(1)
    new_text = new_text.replace('\r', '')
    new_text = re.sub(r'\n*$', '', new_text)
    return new_text + '\n'


def file_is_generated(text):
    if '// This file is automatically generated by scripts/' in text:
        return True
    return False


def format_file(f, full_path, directory, ext):
    global difference_files
    with open_utf8(full_path, 'r') as f:
        old_text = f.read()
    # do not format auto-generated files
    if file_is_generated(old_text) and ext != '.py':
        return
    old_lines = old_text.split('\n')

    new_text = get_formatted_text(f, full_path, directory, ext)
    if ext in ('.cpp', '.hpp'):
        new_text = new_text.replace('ARGS &&...args', 'ARGS &&... args')
    if check_only:
        new_lines = new_text.split('\n')
        old_lines = [x for x in old_lines if '...' not in x]
        new_lines = [x for x in new_lines if '...' not in x]
        diff_result = difflib.unified_diff(old_lines, new_lines)
        total_diff = ""
        for diff_line in diff_result:
            total_diff += diff_line + "\n"
        total_diff = total_diff.strip()

        if len(total_diff) > 0:
            print("----------------------------------------")
            print("----------------------------------------")
            print("Found differences in file " + full_path)
            print("----------------------------------------")
            print("----------------------------------------")
            print(total_diff)
            difference_files.append(full_path)
    else:
        tmpfile = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        with open_utf8(tmpfile, 'w+', newline='\n') as f:
            f.write(new_text)
        shutil.move(tmpfile, full_path)


class ToFormatFile:
    def __init__(self, filename, full_path, directory):
        self.filename = filename
        self.full_path = full_path
        self.directory = directory
        self.ext = '.' + filename.split('.')[-1]


def format_directory(directory):
    files = os.listdir(directory)
    files.sort()
    result = []
    for f in files:
        full_path = os.path.join(directory, f)
        if os.path.isdir(full_path):
            if f in ignored_directories or full_path in ignored_directories:
                continue
            result += format_directory(full_path)
        elif can_format_file(full_path):
            result += [ToFormatFile(f, full_path, directory)]
    return result


files = []
if format_all:
    try:
        os.system(cmake_format_command.replace("${FILE}", "CMakeLists.txt"))
    except:
        pass

    for direct in formatted_directories:
        files += format_directory(direct)

else:
    for full_path in changed_files:
        splits = full_path.split(os.path.sep)
        fname = splits[-1]
        dirname = os.path.sep.join(splits[:-1])
        files.append(ToFormatFile(fname, full_path, dirname))


def process_file(f):
    if not silent:
        print(f.full_path)
    try:
        format_file(f.filename, f.full_path, f.directory, f.ext)
    except:
        print(traceback.format_exc())
        sys.exit(1)


# Create thread for each file
with concurrent.futures.ThreadPoolExecutor() as executor:
    try:
        threads = [executor.submit(process_file, f) for f in files]
        # Wait for all tasks to complete
        concurrent.futures.wait(threads)
    except KeyboardInterrupt:
        executor.shutdown(wait=True, cancel_futures=True)
        raise

if check_only:
    if len(difference_files) > 0:
        print("")
        print("")
        print("")
        print("Failed format-check: differences were found in the following files:")
        for fname in difference_files:
            print("- " + fname)
        print('Run "make format-fix" to fix these differences automatically')
        exit(1)
    else:
        print("Passed format-check")
        exit(0)



# fmt: off

import pytest
import subprocess
import sys
from typing import List
from conftest import ShellTest
import os
from pathlib import Path


# test .open with non-databsae files
def test_open(shell, tmp_path):
    target_dir = tmp_path / 'open_test'
    os.mkdir(target_dir)
    test = (
        ShellTest(shell)
        .statement(f"ATTACH '{target_dir}/test.duckdb' as test")
        .statement("USE test")
        .statement("CREATE TABLE tbl AS SELECT 42")
        .statement("USE memory")
        .statement("DETACH test")
        .statement(f"COPY (SELECT 43) TO '{target_dir}/test.parquet'")
        .statement(f"COPY (SELECT 44) TO '{target_dir}/test.csv'")
        .statement(f"ATTACH '{target_dir}/test.parquet' as test2")
        .statement(f"ATTACH '{target_dir}/test.csv' as test3")
        .statement(f".open '{target_dir}/test.duckdb'")
        .statement("FROM tbl")
        .statement(f".open '{target_dir}/test.parquet'")
        .statement("FROM file")
        .statement(f".open '{target_dir}/test.csv'")
        .statement("FROM file")
    )

    result = test.run()
    result.check_stdout("42")
    result.check_stdout("43")
    result.check_stdout("44")



# fmt: off

from conftest import ShellTest


def test_many_schemas(shell):
    test = (
        ShellTest(shell)
        .statement('''
attach ':memory:' as my_db1;
attach ':memory:' as my_db2;
create schema my_db1.db1_s1;
create schema my_db1.db1_s2;
create schema my_db2.db2_s1;
create schema my_db2.db2_s2;
create schema memory.memory_s1;
create schema memory.memory_s2;''')
        .statement('show schemas')
    )

    result = test.run()
    result.check_stdout("memory")
    result.check_stdout("my_db1")
    result.check_stdout("my_db2")
    result.check_stdout("main")
    result.check_stdout("memory_s1")
    result.check_stdout("memory_s2")
    result.check_stdout("db1_s1")
    result.check_stdout("db1_s2")
    result.check_stdout("db2_s1")
    result.check_stdout("db2_s2")


def test_detaching_default_db(shell):
    test = (
        ShellTest(shell)
        .statement("attach ':memory:' as new_db")
        .statement("use new_db")
        .statement("detach memory")
        .statement("show schemas")
    )

    result = test.run()
    result.check_stdout("new_db")
    result.check_stdout("main")
    result.check_stdout("true")
    result.check_not_exist("memory")



import subprocess
import time
import argparse
import os

parser = argparse.ArgumentParser(description='Runs the memory leak tests')

parser.add_argument(
    '--unittest',
    dest='unittest',
    action='store',
    help='Path to unittest executable',
    default='build/release/test/unittest',
)
parser.add_argument('--test', dest='test', action='store', help='The name of the tests to run (* is all)', default='*')
parser.add_argument(
    '--timeout',
    dest='timeout',
    action='store',
    help='The maximum time to run the test and measure memory usage (in seconds)',
    default=60,
)
parser.add_argument(
    '--threshold-percentage',
    dest='threshold_percentage',
    action='store',
    help='The percentage threshold before we consider an increase a regression',
    default=0.01,
)
parser.add_argument(
    '--threshold-absolute',
    dest='threshold_absolute',
    action='store',
    help='The absolute threshold before we consider an increase a regression',
    default=1000,
)
parser.add_argument('--verbose', dest='verbose', action='store', help='Verbose output', default=True)

args = parser.parse_args()

unittest_program = args.unittest
test_filter = args.test
test_time = int(args.timeout)
verbose = args.verbose
measurements_per_second = 1.0

# get a list of all unittests
proc = subprocess.Popen([unittest_program, '-l', '[memoryleak]'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
stdout = proc.stdout.read().decode('utf8')
stderr = proc.stderr.read().decode('utf8')
if proc.returncode is not None and proc.returncode != 0:
    print("Failed to run program " + unittest_program)
    print(proc.returncode)
    print(stdout)
    print(stderr)
    exit(1)

test_cases = []
first_line = True
for line in stdout.splitlines():
    if first_line:
        first_line = False
        continue
    if len(line.strip()) == 0:
        continue
    splits = line.rsplit('\t', 1)
    if test_filter == '*' or test_filter in splits[0]:
        test_cases.append(splits[0])

if len(test_cases) == 0:
    print(f"No tests matching filter \"{test_filter}\" found")
    exit(0)


def sizeof_fmt(num, suffix="B"):
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1000.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1000.0
    return f"{num:.1f}Yi{suffix}"


def run_test(test_case):
    # launch the unittest program
    proc = subprocess.Popen(
        [unittest_program, test_case, '--memory-leak-tests'], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    pid = proc.pid

    # capture the memory output for the duration of the program running
    leak = True
    rss = []
    for i in range(int(test_time * measurements_per_second)):
        time.sleep(1.0 / measurements_per_second)
        if proc.poll() is not None:
            print("------------------------------------------------")
            print("                    FAILURE                     ")
            print("------------------------------------------------")
            print("                    stdout:                     ")
            print("------------------------------------------------")
            print(proc.stdout.read().decode('utf8'))
            print("------------------------------------------------")
            print("                    stderr:                     ")
            print("------------------------------------------------")
            print(proc.stderr.read().decode('utf8'))
            exit(1)
        ps_proc = subprocess.Popen(f'ps -o rss= -p {pid}'.split(' '), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        res = ps_proc.stdout.read().decode('utf8').strip()
        memory_usage_in_bytes = int(res) * 1024
        if verbose:
            print(f"{i / measurements_per_second}: {sizeof_fmt(memory_usage_in_bytes)}")

        rss.append(memory_usage_in_bytes)
        if not has_memory_leak(rss):
            leak = False
            break

    proc.terminate()
    if leak:
        print("------------------------------------------------")
        print("                     ERROR                      ")
        print("------------------------------------------------")
        print(f"Memory leak detected in test case \"{test_case}\"")
        print("------------------------------------------------")
    elif verbose:
        print("------------------------------------------------")
        print("                    Success!                    ")
        print("------------------------------------------------")
        print("------------------------------------------------")
        print(f"No memory leaks detected in test case \"{test_case}\"")
        print("------------------------------------------------")
    if verbose or leak:
        print("Observed memory usages")
        print("------------------------------------------------")
        for i in range(len(rss)):
            memory = rss[i]
            print(f"{i / measurements_per_second}: {sizeof_fmt(memory)}")
        if leak:
            exit(1)


def has_memory_leak(rss):
    measurement_count = int(measurements_per_second * 10)
    if len(rss) <= measurement_count:
        # not enough measurements yet
        return True
    differences = []
    for i in range(1, len(rss)):
        differences.append(rss[i] - rss[i - 1])
    max_memory = max(rss)
    sum_differences = sum(differences[-measurement_count:])
    return sum_differences > (max_memory * args.threshold_percentage + args.threshold_absolute)


try:
    for index, test in enumerate(test_cases):
        print(f"[{index}/{len(test_cases)}] {test}")
        run_test(test)
finally:
    os.system('killall -9 unittest')



import argparse
import os
import duckdb_sqllogictest
from duckdb_sqllogictest import SQLParserException, SQLLogicParser, SQLLogicTest
import subprocess
import multiprocessing
import tempfile
import re

parser = argparse.ArgumentParser(description="Test serialization")
parser.add_argument("--shell", type=str, help="Shell binary to run", default=os.path.join('build', 'debug', 'duckdb'))
parser.add_argument("--offset", type=int, help="File offset", default=None)
parser.add_argument("--count", type=int, help="File count", default=None)
parser.add_argument('--no-exit', action='store_true', help='Do not exit after a test fails', default=False)
parser.add_argument('--print-failing-only', action='store_true', help='Print failing tests only', default=False)
parser.add_argument(
    '--include-extensions', action='store_true', help='Include test files of out-of-tree extensions', default=False
)
group = parser.add_mutually_exclusive_group(required=True)
group.add_argument("--test-file", type=str, help="Path to the SQL logic file", default='')
group.add_argument(
    "--test-list", type=str, help="Path to the file that contains a newline separated list of test files", default=''
)
group.add_argument("--all-tests", action='store_true', help="Run all tests", default=False)
args = parser.parse_args()


def extract_git_urls(script: str):
    pattern = r'GIT_URL\s+(https?://\S+)'
    return re.findall(pattern, script)


import os
import requests
from urllib.parse import urlparse


def download_directory_contents(api_url, local_path, headers):
    response = requests.get(api_url, headers=headers)
    if response.status_code != 200:
        print(f"⚠️  Could not access {api_url}: {response.status_code}")
        return

    os.makedirs(local_path, exist_ok=True)

    for item in response.json():
        item_type = item.get("type")
        item_name = item.get("name")
        if item_type == "file":
            download_url = item.get("download_url")
            if not download_url:
                continue
            file_path = os.path.join(local_path, item_name)
            file_resp = requests.get(download_url)
            if file_resp.status_code == 200:
                with open(file_path, "wb") as f:
                    f.write(file_resp.content)
                print(f"  - Downloaded {file_path}")
            else:
                print(f"  - Failed to download {file_path}")
        elif item_type == "dir":
            subdir_api_url = item.get("url")
            subdir_local_path = os.path.join(local_path, item_name)
            download_directory_contents(subdir_api_url, subdir_local_path, headers)


def download_test_sql_folder(repo_url, base_folder="extension-test-files"):
    repo_name = urlparse(repo_url).path.strip("/").split("/")[-1]
    target_folder = os.path.join(base_folder, repo_name)

    if os.path.exists(target_folder):
        print(f"✓ Skipping {repo_name}, already exists.")
        return

    print(f"⬇️ Downloading test/sql from {repo_name}...")

    api_url = f"https://api.github.com/repos/duckdb/{repo_name}/contents/test/sql?ref=main"
    GITHUB_TOKEN = os.environ["GITHUB_TOKEN"]
    headers = {"Accept": "application/vnd.github.v3+json", "Authorization": f"Bearer {GITHUB_TOKEN}"}

    download_directory_contents(api_url, target_folder, headers)


def batch_download_all_test_sql():
    filename = ".github/config/out_of_tree_extensions.cmake"
    if not os.path.isfile(filename):
        raise Exception(f"File {filename} not found")
    with open(filename, "r") as f:
        content = f.read()
    urls = extract_git_urls(content)
    if urls == []:
        print("No URLs found.")
    for url in urls:
        download_test_sql_folder(url)


def find_tests_recursive(dir, excluded_paths):
    test_list = []
    for f in os.listdir(dir):
        path = os.path.join(dir, f)
        if path in excluded_paths:
            continue
        if os.path.isdir(path):
            test_list += find_tests_recursive(path, excluded_paths)
        elif path.endswith('.test') or path.endswith('.test_slow'):
            test_list.append(path)
    return test_list


def parse_test_file(filename):
    if not os.path.isfile(filename):
        raise Exception(f"File {filename} not found")
    parser = SQLLogicParser()
    try:
        out: Optional[SQLLogicTest] = parser.parse(filename)
        if not out:
            raise SQLParserException(f"Test {filename} could not be parsed")
    except:
        return []
    loop_count = 0
    statements = []
    for stmt in out.statements:
        if type(stmt) is duckdb_sqllogictest.statement.skip.Skip:
            # mode skip - just skip entire test
            break
        if (
            type(stmt) is duckdb_sqllogictest.statement.loop.Loop
            or type(stmt) is duckdb_sqllogictest.statement.foreach.Foreach
        ):
            loop_count += 1
        if type(stmt) is duckdb_sqllogictest.statement.endloop.Endloop:
            loop_count -= 1
        if loop_count > 0:
            # loops are ignored currently
            continue
        if not (
            type(stmt) is duckdb_sqllogictest.statement.query.Query
            or type(stmt) is duckdb_sqllogictest.statement.statement.Statement
        ):
            # only handle query and statement nodes for now
            continue
        if type(stmt) is duckdb_sqllogictest.statement.Statement:
            # skip expected errors
            if stmt.expected_result.type == duckdb_sqllogictest.ExpectedResult.Type.ERROR:
                if any(
                    "parser error" in line.lower() or "syntax error" in line.lower()
                    for line in stmt.expected_result.lines
                ):
                    continue
        query = ' '.join(stmt.lines)
        statements.append(query)
    return statements


def run_test_case(args_tuple):
    i, file, shell, print_failing_only = args_tuple
    results = []
    if not print_failing_only:
        print(f"Run test {i}: {file}")

    statements = parse_test_file(file)
    for statement in statements:
        with tempfile.TemporaryDirectory() as tmpdir:
            peg_sql_path = os.path.join(tmpdir, 'peg_test.sql')
            with open(peg_sql_path, 'w') as f:
                f.write(f'CALL check_peg_parser($TEST_PEG_PARSER${statement}$TEST_PEG_PARSER$);\n')

            proc = subprocess.run([shell, '-init', peg_sql_path, '-c', '.exit'], capture_output=True)
            stderr = proc.stderr.decode('utf8')

            if proc.returncode == 0 and ' Error:' not in stderr:
                continue

            if print_failing_only:
                print(f"Failed test {i}: {file}")
            else:
                print(f'Failed')
                print(f'-- STDOUT --')
                print(proc.stdout.decode('utf8'))
                print(f'-- STDERR --')
                print(stderr)

            results.append((file, statement))
            break
    return results


if __name__ == "__main__":
    files = []
    excluded_tests = {
        'test/sql/peg_parser',  # Fail for some reason
        'test/sql/prepared/parameter_variants.test',  # PostgreSQL parser bug with ?1
        'test/sql/copy/s3/download_config.test',  # Unknown why this passes in SQLLogicTest
        'test/sql/function/list/lambdas/arrow/lambda_scope_deprecated.test',  # Error in the tokenization of *+*
        'test/sql/catalog/function/test_simple_macro.test',  # Bug when mixing named parameters and non-named
    }
    if args.all_tests:
        # run all tests
        test_dir = os.path.join('test', 'sql')
        files = find_tests_recursive(test_dir, excluded_tests)
        if args.include_extensions:
            batch_download_all_test_sql()
            extension_files = find_tests_recursive('extension-test-files', {})
            files = files + extension_files
    elif len(args.test_list) > 0:
        with open(args.test_list, 'r') as f:
            files = [x.strip() for x in f.readlines() if x.strip() not in excluded_tests]
    else:
        # run a single test
        files.append(args.test_file)
    files.sort()

    start = args.offset if args.offset is not None else 0
    end = start + args.count if args.count is not None else len(files)
    work_items = [(i, files[i], args.shell, args.print_failing_only) for i in range(start, end)]

    if not args.no_exit:
        # Disable multiprocessing for --no-exit behavior
        failed_test_list = []
        for item in work_items:
            res = run_test_case(item)
            if res:
                failed_test_list.extend(res)
                exit(1)
    else:
        with multiprocessing.Pool() as pool:
            results = pool.map(run_test_case, work_items)
        failed_test_list = [item for sublist in results for item in sublist]

    failed_tests = len(failed_test_list)
    print("List of failed tests: ")
    for test, statement in failed_test_list:
        print(f"{test}\n{statement}\n\n")
    print(f"Total of {failed_tests} out of {len(files)} failed ({round(failed_tests/len(files) * 100,2)}%). ")



from __future__ import annotations
import json
import re
from pathlib import Path

IDENT_RE = re.compile(r"^[A-Z_][A-Z0-9_]*$")

START_OF_FILE = "// DUCKDB_START_OF_FILE"
INSERT_CODE_HERE = "// DUCKDB_INSERT_CODE_HERE"


def load_metrics_json(path: Path) -> list[dict]:
    if not path.exists():
        raise FileNotFoundError(f"metric_type.json not found at {path}")
    with path.open("r", encoding="utf-8") as f:
        return json.load(f)


def validate_identifier(name: str, group: str) -> None:
    if not IDENT_RE.match(name):
        raise ValueError(f"Invalid metric identifier: {name}, in group {group}")


def _to_pascal_case(s: str) -> str:
    return ''.join(word.capitalize() for word in s.split('_'))


def retrieve_optimizers(optimizer_file: Path) -> list[tuple[str, int]]:
    if not optimizer_file.exists():
        raise FileNotFoundError(f"optimizer_type.hpp not found at {optimizer_file}.")
    enum_with_value = re.compile(r"\s*([A-Z_]+)\s*=\s*(\d+)\s*,?")
    enum_without_value = re.compile(r"\s*([A-Z_]+)\s*,?")
    inside_enum = False
    result: list[tuple[str, int]] = []
    with optimizer_file.open("r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if line.startswith("enum class OptimizerType"):
                inside_enum = True
                continue

            if inside_enum and line.startswith("};"):
                break

            if inside_enum:
                m = enum_with_value.match(line)
                if m:
                    name = m[1]
                    value = int(m[2])
                    if name == "INVALID":
                        continue
                    result.append((name, value))
                    continue
                m = enum_without_value.match(line)
                if m:
                    name = m[1]
                    if name == "INVALID":
                        continue
                    raise ValueError(
                        f"OptimizerType entry '{name}' has no explicit value. "
                        f"All OptimizerType entries must have explicit values for backwards compatibility."
                    )

    return result


def retrieve_template(profiling_util_file: Path, start: str = START_OF_FILE, end: str = "EOF") -> str:
    if not profiling_util_file.exists():
        raise FileNotFoundError(f"file not found at {profiling_util_file}.")

    with profiling_util_file.open("r", encoding="utf-8") as f:
        result = f.read()

    if start not in result:
        print(f"Could not find the start of file mark: {start}")

    if end != "EOF":
        result = result[: result.find(end)]
        return result.split(start)[1].split(end)[0]

    return result.split(start)[1]



from settings_scripts import parse_and_sort_json_file, update_header_file, update_scopes, update_src_code
from settings_scripts.config import SettingsList, make_format

if __name__ == '__main__':
    parse_and_sort_json_file()
    update_header_file()
    update_scopes()
    update_src_code()
    make_format()
    print(f"- Successfully parsed and included {len(SettingsList)} setting(s)!")



import os
import re
import subprocess
import tempfile
from pathlib import Path
from typing import Set, List
from functools import total_ordering

# define file paths and global variables
DUCKDB_DIR = Path(__file__).resolve().parent.parent.parent
DUCKDB_SETTINGS_HEADER_FILE = os.path.join(DUCKDB_DIR, "src/include/duckdb/main", "settings.hpp")
DUCKDB_AUTOGENERATED_SETTINGS_FILE = os.path.join(DUCKDB_DIR, "src/main/settings", "autogenerated_settings.cpp")
DUCKDB_SETTINGS_SCOPE_FILE = os.path.join(DUCKDB_DIR, "src/main", "config.cpp")
JSON_PATH = os.path.join(DUCKDB_DIR, "src/common", "settings.json")

# define scope values
VALID_SCOPE_VALUES = ["GLOBAL", "LOCAL", "GLOBAL_LOCAL"]
INVALID_SCOPE_VALUE = "INVALID"
SQL_TYPE_MAP = {
    "UBIGINT": "idx_t",
    "BIGINT": "int64_t",
    "BOOLEAN": "bool",
    "DOUBLE": "double",
    "VARCHAR": "string",
}

setting_index = 0


# global Setting structure
@total_ordering
class Setting:
    # track names of written settings to prevent duplicates
    __written_settings: Set[str] = set()

    def __init__(
        self,
        name: str,
        description: str,
        sql_type: str,
        scope: str,
        internal_setting: str,
        on_callbacks: List[str],
        custom_implementation,
        struct_name: str,
        aliases: List[str],
        default_scope: str,
        default_value: str,
        conditional_defaults,
    ):
        self.name = self._get_valid_name(name)
        self.description = description
        self.sql_type = self._get_sql_type(sql_type)
        self.return_type = self._get_setting_type(sql_type)
        self.is_enum = sql_type.startswith('ENUM')
        self.internal_setting = internal_setting
        self.scope = self._get_valid_scope(scope) if scope is not None else None
        self.on_set, self.on_reset = self._get_on_callbacks(on_callbacks)
        self.is_generic_setting = default_value is not None
        self.setting_index = None
        if self.is_enum and self.is_generic_setting:
            self.on_set = True
        custom_callbacks = ['set', 'reset', 'get']
        if type(custom_implementation) is bool:
            self.all_custom = custom_implementation
            self.custom_implementation = custom_callbacks if custom_implementation else []
        else:
            for entry in custom_implementation:
                if entry not in custom_callbacks:
                    raise ValueError(
                        f"Setting {self.name} - incorrect input for custom_implementation - expected set/reset/get, got {entry}"
                    )
            self.all_custom = len(set(custom_implementation)) == 3
            self.custom_implementation = custom_implementation
        self.aliases = self._get_aliases(aliases)
        self.struct_name = self._get_struct_name() if len(struct_name) == 0 else struct_name
        self.default_scope = self._get_valid_default_scope(default_scope) if default_scope is not None else None
        self.default_value = default_value
        self.conditional_defaults = conditional_defaults
        if self.default_value is not None:
            global setting_index
            self.setting_index = setting_index
            setting_index += 1

        if self.default_scope is not None and self.scope is not None:
            raise ValueError("Only default_scope or scope can be specified")

    # define all comparisons to be based on the setting's name attribute
    def __eq__(self, other) -> bool:
        return isinstance(other, Setting) and self.name == other.name

    def __lt__(self, other) -> bool:
        return isinstance(other, Setting) and self.name < other.name

    def __hash__(self) -> int:
        return hash(self.name)

    def __repr__(self):
        return f"struct {self.struct_name} -> {self.name}, {self.sql_type}, {self.type}, {self.scope}, {self.description} {self.aliases}"

    # validate setting name for correct format and uniqueness
    def _get_valid_name(self, name: str) -> str:
        if not re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', name):
            raise ValueError(f"'{name}' cannot be used as setting name - invalid character")
        if name in Setting.__written_settings:
            raise ValueError(f"'{name}' cannot be used as setting name - already exists")
        Setting.__written_settings.add(name)
        return name

    # ensure the setting scope is valid based on the accepted values
    def _get_valid_scope(self, scope: str) -> str:
        scope = scope.upper()
        if scope in VALID_SCOPE_VALUES:
            return scope
        return INVALID_SCOPE_VALUE

    def _get_valid_default_scope(self, scope: str) -> str:
        scope = scope.upper()
        if scope == 'GLOBAL':
            return scope
        elif scope == 'LOCAL':
            return 'LOCAL'
        raise Exception(f"Invalid default scope value {scope}")

    # validate and return the correct type format
    def _get_sql_type(self, sql_type) -> str:
        if sql_type.startswith('ENUM'):
            return 'VARCHAR'
        if sql_type.endswith('[]'):
            # recurse into child-element
            sub_type = self._get_sql_type(sql_type[:-2])
            return sql_type
        if sql_type in SQL_TYPE_MAP:
            return sql_type
        raise ValueError(f"Invalid SQL type: '{sql_type}' - supported types are {', '.join(SQL_TYPE_MAP.keys())}")

    # validate and return the cpp input type
    def _get_setting_type(self, type) -> str:
        if type.startswith('ENUM'):
            return type[len('ENUM<') : -1]
        if type.endswith('[]'):
            subtype = self._get_setting_type(type[:-2])
            return "vector<" + subtype + ">"
        return SQL_TYPE_MAP[type]

    # validate and return the correct type format
    def _get_on_callbacks(self, callbacks) -> (bool, bool):
        set = False
        reset = False
        for entry in callbacks:
            if entry == 'set':
                set = True
            elif entry == 'reset':
                reset = True
            else:
                raise ValueError(f"Invalid entry in on_callbacks list: {entry} (expected set or reset)")
        return (set, reset)

    # validate and return the set of the aliases
    def _get_aliases(self, aliases: List[str]) -> List[str]:
        return [self._get_valid_name(alias) for alias in aliases]

    # generate a function name
    def _get_struct_name(self) -> str:
        camel_case_name = ''.join(word.capitalize() for word in re.split(r'[-_]', self.name))
        if camel_case_name.endswith("Setting"):
            return f"{camel_case_name}"
        return f"{camel_case_name}Setting"


# this global list (accessible across all files) stores all the settings definitions in the json file
SettingsList: List[Setting] = []


# global method that finds the indexes of a start and an end marker in a file
def find_start_end_indexes(source_code, start_marker, end_marker, file_path):
    start_matches = list(re.finditer(start_marker, source_code))
    if len(start_matches) == 0:
        raise ValueError(f"Couldn't find start marker {start_marker} in {file_path}")
    elif len(start_matches) > 1:
        raise ValueError(f"Start marker found more than once in {file_path}")
    start_index = start_matches[0].end()

    end_matches = list(re.finditer(end_marker, source_code[start_index:]))
    if len(end_matches) == 0:
        raise ValueError(f"Couldn't find end marker {end_marker} in {file_path}")
    elif len(end_matches) > 1:
        raise ValueError(f"End marker found more than once in {file_path}")
    end_index = start_index + end_matches[0].start()
    return start_index, end_index


# global markers
SEPARATOR = "//===----------------------------------------------------------------------===//\n"
SRC_CODE_START_MARKER = "namespace duckdb {"
SRC_CODE_END_MARKER = "} // namespace duckdb"


# global method
def write_content_to_file(new_content, path):
    with open(path, 'w') as source_file:
        source_file.write("".join(new_content))


def get_setting_heading(setting_struct_name):
    struct_name_wt_Setting = re.sub(r'Setting$', '', setting_struct_name)
    heading_name = re.sub(r'(?<!^)(?=[A-Z])', ' ', struct_name_wt_Setting)
    heading = SEPARATOR + f"// {heading_name}\n" + SEPARATOR
    return heading


def make_format():
    os.system(f"python3 scripts/format.py {DUCKDB_SETTINGS_HEADER_FILE} --fix --force --noconfirm")
    os.system(f"python3 scripts/format.py {DUCKDB_SETTINGS_SCOPE_FILE} --fix --force --noconfirm")
    os.system(f"python3 scripts/format.py {DUCKDB_AUTOGENERATED_SETTINGS_FILE} --fix --force --noconfirm")



import os
import json
import re
import argparse
from enum import Enum

from typing import Dict, Optional, Tuple, List

parser = argparse.ArgumentParser(description='Generate serialization code')
parser.add_argument('--source', type=str, help='Source directory')
parser.add_argument('--target', type=str, help='Target directory')

args = parser.parse_args()


class MemberVariableStatus(Enum):
    # Both serialized and deserialized
    EXISTING = 1
    # Not serialized, but is deserialized
    READ_ONLY = 2
    # Not serialized, not deserialized
    DELETED = 3


def get_file_list():
    if args.source is None:
        targets = [
            {'source': 'src/include/duckdb/storage/serialization', 'target': 'src/storage/serialization'},
            {'source': 'extension/parquet/include/', 'target': 'extension/parquet'},
            {'source': 'extension/json/include/', 'target': 'extension/json'},
        ]
    else:
        targets = [
            {'source': args.source, 'target': args.target},
        ]

    file_list = []
    for target in targets:
        source_base = os.path.sep.join(target['source'].split('/'))
        target_base = os.path.sep.join(target['target'].split('/'))
        for fname in os.listdir(source_base):
            if '.json' not in fname:
                continue
            if '_enums.json' in fname:
                continue
            file_list.append(
                {
                    'source': os.path.join(source_base, fname),
                    'target': os.path.join(target_base, 'serialize_' + fname.replace('.json', '.cpp')),
                }
            )
    return file_list


scripts_dir = os.path.dirname(os.path.abspath(__file__))
version_map_path = os.path.join(scripts_dir, '..', 'src', 'storage', 'version_map.json')
version_map_file = file = open(version_map_path)
version_map = json.load(version_map_file)


def verify_serialization_versions(version_map):
    serialization = version_map['serialization']['values']
    if list(serialization.keys())[-1] != 'latest':
        print(f"The version map ({version_map_path}) for serialization versions must end in 'latest'!")
        exit(1)


verify_serialization_versions(version_map)


def lookup_serialization_version(version: str):
    if version.lower() == "latest":
        print(
            f"'latest' is not an allowed 'version' to use in serialization JSON files, please provide a duckdb version"
        )

    versions = version_map['serialization']['values']
    if version not in versions:
        from packaging.version import Version

        current_version = Version(version)

        # This version does not exist in the version map
        # Which is allowed for unreleased versions, they will get mapped to 'latest' instead

        last_registered_version = Version(list(versions.keys())[-2])
        if current_version < last_registered_version:
            # The version was lower than the last defined version, which is not allowed
            print(
                f"Specified version ({current_version}) could not be found in the version_map.json, and it is lower than the last defined version ({last_registered_version})!"
            )
            exit(1)
        if hasattr(versions, 'latest'):
            # We have already mapped a version to 'latest', check that the versions match
            latest_version = getattr(versions, 'latest')
            if current_version != latest_version:
                print(
                    f"Found more than one version that is not present in the version_map.json!: Current: {current_version}, Latest: {latest_version}"
                )
                exit(1)
        else:
            setattr(lookup_serialization_version, 'latest', current_version)
        return versions['latest']
    return versions[version]


INCLUDE_FORMAT = '#include "{filename}"\n'

HEADER = '''//===----------------------------------------------------------------------===//
// This file is automatically generated by scripts/generate_serialization.py
// Do not edit this file manually, your changes will be overwritten
//===----------------------------------------------------------------------===//

{include_list}
namespace duckdb {{
'''

FOOTER = '''
} // namespace duckdb
'''

TEMPLATED_BASE_FORMAT = '''
template <typename {template_name}>'''

SERIALIZE_BASE_FORMAT = '''
void {class_name}::Serialize(Serializer &serializer) const {{
{members}}}
'''

SERIALIZE_ELEMENT_FORMAT = (
    '\tserializer.WriteProperty<{property_type}>({property_id}, "{property_key}", {property_name}{property_default});\n'
)

BASE_SERIALIZE_FORMAT = '\t{base_class_name}::Serialize(serializer);\n'

POINTER_RETURN_FORMAT = '{pointer}<{class_name}>'

DESERIALIZE_BASE_FORMAT = '''
{deserialize_return} {class_name}::Deserialize(Deserializer &deserializer) {{
{members}
}}
'''

SWITCH_CODE_FORMAT = '''\tswitch ({switch_variable}) {{
{case_statements}\tdefault:
\t\tthrow SerializationException("Unsupported type for deserialization of {base_class}!");
\t}}
'''

SET_DESERIALIZE_PARAMETER_FORMAT = '\tdeserializer.Set<{property_type}>({property_name});\n'
UNSET_DESERIALIZE_PARAMETER_FORMAT = '\tdeserializer.Unset<{property_type}>();\n'
GET_DESERIALIZE_PARAMETER_FORMAT = 'deserializer.Get<{property_type}>()'
TRY_GET_DESERIALIZE_PARAMETER_FORMAT = 'deserializer.TryGet<{property_type}>()'

SWITCH_HEADER_FORMAT = '\tcase {enum_type}::{enum_value}:\n'

SWITCH_STATEMENT_FORMAT = (
    SWITCH_HEADER_FORMAT
    + '''\t\tresult = {class_deserialize}::Deserialize(deserializer);
\t\tbreak;
'''
)

DESERIALIZE_ELEMENT_FORMAT = '\tauto {property_name} = deserializer.ReadProperty<{property_type}>({property_id}, "{property_key}"{property_default});\n'
DESERIALIZE_ELEMENT_BASE_FORMAT = '\tauto {property_name} = deserializer.ReadProperty<unique_ptr<{base_property}>>({property_id}, "{property_key}"{property_default});\n'
DESERIALIZE_ELEMENT_CLASS_FORMAT = '\tdeserializer.ReadProperty<{property_type}>({property_id}, "{property_key}", result{assignment}{property_name}{property_default});\n'
DESERIALIZE_ELEMENT_CLASS_BASE_FORMAT = '\tauto {property_name} = deserializer.ReadProperty<unique_ptr<{base_property}>>({property_id}, "{property_key}"{property_default});\n\tresult{assignment}{property_name} = unique_ptr_cast<{base_property}, {derived_property}>(std::move({property_name}));\n'

MOVE_LIST = [
    'string',
    'ParsedExpression*',
    'CommonTableExpressionMap',
    'LogicalType',
    'ColumnDefinition',
    'BaseStatistics',
    'BoundLimitNode',
]

REFERENCE_LIST = ['ClientContext', 'bound_parameter_map_t', 'Catalog']


def is_container(type):
    return '<' in type and 'CSVOption' not in type


def is_pointer(type):
    return type.endswith('*') or type.startswith('shared_ptr<')


def is_zeroable(type):
    return type in [
        'bool',
        'int8_t',
        'int16_t',
        'int32_t',
        'int64_t',
        'uint8_t',
        'uint16_t',
        'uint32_t',
        'uint64_t',
        'idx_t',
        'size_t',
        'int',
    ]


def requires_move(type):
    return is_container(type) or is_pointer(type) or type in MOVE_LIST


def replace_pointer(type):
    return re.sub('([a-zA-Z0-9]+)[*]', 'unique_ptr<\\1>', type)


def get_default_argument(default_value):
    return f'{default_value}'.lower() if type(default_value) == bool else f'{default_value}'


def get_deserialize_element_template(
    template,
    property_name,
    property_key,
    property_id,
    property_type,
    has_default,
    default_value,
    status: MemberVariableStatus,
    pointer_type,
):
    if status == MemberVariableStatus.READ_ONLY and not has_default:
        print("'read_only' status is not allowed without a default value")
        exit(1)

    # read_method = 'ReadProperty'
    assignment = '.' if pointer_type == 'none' else '->'
    default_argument = '' if default_value is None else f', {get_default_argument(default_value)}'
    if status == MemberVariableStatus.DELETED:
        template = template.replace(', result{assignment}{property_name}', '').replace(
            'ReadProperty', 'ReadDeletedProperty'
        )
    elif has_default and default_value is None:
        template = template.replace('ReadProperty', 'ReadPropertyWithDefault')
    elif has_default and default_value is not None:
        template = template.replace('ReadProperty', 'ReadPropertyWithExplicitDefault')
    template = template.format(
        property_name=property_name,
        property_key=property_key,
        property_id=str(property_id),
        property_default=default_argument,
        property_type=property_type,
        assignment=assignment,
    )
    if status == MemberVariableStatus.DELETED:
        template = template.replace(f'auto {property_name} = ', '')
    return template


def get_deserialize_assignment(property_name, property_type, pointer_type):
    assignment = '.' if pointer_type == 'none' else '->'
    property = property_name.replace('.', '_')
    if requires_move(property_type):
        property = f'std::move({property})'
    return f'\tresult{assignment}{property_name} = {property};\n'


def get_return_value(pointer_type, class_name):
    if pointer_type == 'none':
        return class_name
    return POINTER_RETURN_FORMAT.format(pointer=pointer_type, class_name=class_name)


def generate_return(class_entry):
    if class_entry.base is None or class_entry.constructor_method is not None:
        return '\treturn result;'
    else:
        return '\treturn std::move(result);'


def parse_status(status: str):
    if status == 'deleted':
        return MemberVariableStatus.DELETED
    if status == 'read_only':
        return MemberVariableStatus.READ_ONLY
    if status == 'existing':
        return MemberVariableStatus.EXISTING
    valid_options = ['deleted', 'read_only', 'existing']
    valid_options_string = ", ".join(valid_options)
    print(f"Invalid 'status' ('{status}') encountered, valid options are: {valid_options_string}")
    exit(1)


# FIXME: python has __slots__ for this, so it's enforced by Python itself
# see: https://wiki.python.org/moin/UsingSlots
supported_member_entries = [
    'id',
    'name',
    'type',
    'property',
    'serialize_property',
    'deserialize_property',
    'base',
    'default',
    'status',
    'version',
]


def has_default_by_default(type):
    if is_pointer(type):
        return True
    if is_container(type):
        if 'IndexVector' in type:
            return False
        if 'CSVOption' in type:
            return False
        return True
    if type == 'string':
        return True
    if is_zeroable(type):
        return True
    return False


class MemberVariable:
    def __init__(self, entry):
        self.id = entry['id']
        self.name = entry['name']
        self.type = entry['type']
        self.base = None
        self.has_default = False
        self.default = None
        self.status: MemberVariableStatus = MemberVariableStatus.EXISTING
        self.version: str = 'v0.10.2'
        if 'property' in entry:
            self.serialize_property = entry['property']
            self.deserialize_property = entry['property']
        else:
            self.serialize_property = self.name
            self.deserialize_property = self.name
        if 'version' in entry:
            self.version = entry['version']
        if 'serialize_property' in entry:
            self.serialize_property = entry['serialize_property']
        if 'deserialize_property' in entry:
            self.deserialize_property = entry['deserialize_property']
        if 'default' in entry:
            self.has_default = True
            self.default = entry['default']
        if 'status' in entry:
            self.status = parse_status(entry['status'])
        if self.default is None:
            # default default
            self.has_default = has_default_by_default(self.type)
        if 'base' in entry:
            self.base = entry['base']
        for key in entry.keys():
            if key not in supported_member_entries:
                print(
                    f"Unsupported key \"{key}\" in member variable, key should be in set {str(supported_member_entries)}"
                )


supported_serialize_entries = [
    'class',
    'class_type',
    'pointer_type',
    'base',
    'enum',
    'constructor',
    'constructor_method',
    'custom_implementation',
    'custom_switch_code',
    'members',
    'return_type',
    'set_parameters',
    'includes',
    'finalize_deserialization',
]


class SerializableClass:
    def __init__(self, entry):
        self.name = entry['class']
        self.is_base_class = 'class_type' in entry
        self.base = None
        self.base_object = None
        self.enum_value = None
        self.enum_entries = []
        self.set_parameter_names = []
        self.set_parameters = []
        self.pointer_type = 'unique_ptr'
        self.constructor: Optional[List[str]] = None
        self.constructor_method = None
        self.members: Optional[List[MemberVariable]] = None
        self.custom_implementation = False
        self.custom_switch_code = None
        self.children: Dict[str, SerializableClass] = {}
        self.return_type = self.name
        self.return_class = self.name
        self.finalize_deserialization = None
        if 'finalize_deserialization' in entry:
            self.finalize_deserialization = entry['finalize_deserialization']
        if self.is_base_class:
            self.enum_value = entry['class_type']
        if 'pointer_type' in entry:
            self.pointer_type = entry['pointer_type']
        if 'base' in entry:
            self.base = entry['base']
            self.enum_entries = entry['enum']
            if type(self.enum_entries) is str:
                self.enum_entries = [self.enum_entries]
            self.return_type = self.base
        if 'constructor' in entry:
            self.constructor = entry['constructor']
            if not isinstance(self.constructor, list):
                print(f"constructor for {self.name}, must be of type [], but is of type {str(type(self.constructor))}")
                exit(1)
        if 'constructor_method' in entry:
            self.constructor_method = entry['constructor_method']
            if self.constructor is not None:
                print(
                    "Not allowed to mix 'constructor_method' and 'constructor', 'constructor_method' will implicitly receive all parameters"
                )
                exit(1)
        if 'custom_implementation' in entry and entry['custom_implementation']:
            self.custom_implementation = True
        if 'custom_switch_code' in entry:
            self.custom_switch_code = entry['custom_switch_code']
        if 'members' in entry:
            self.members = [MemberVariable(x) for x in entry['members']]
        if 'return_type' in entry:
            self.return_type = entry['return_type']
            self.return_class = self.return_type
        if 'set_parameters' in entry:
            self.set_parameter_names = entry['set_parameters']
            for set_parameter_name in self.set_parameter_names:
                found = False
                assert self.members is not None
                for member in self.members:
                    if member.name == set_parameter_name:
                        self.set_parameters.append(member)
                        found = True
                        break
                if not found:
                    raise Exception(f'Set parameter {set_parameter_name} not found in member list')
        for key in entry.keys():
            if key not in supported_serialize_entries:
                print(
                    f"Unsupported key \"{key}\" in member variable, key should be in set {str(supported_serialize_entries)}"
                )

    def inherit(self, base_class):
        self.base_object = base_class
        self.pointer_type = base_class.pointer_type

    def get_deserialize_element(
        self, entry: MemberVariable, *, base: Optional[str] = None, pointer_type: Optional[str] = None
    ):
        property_name = entry.deserialize_property
        property_id = entry.id
        property_key = entry.name
        property_type = replace_pointer(entry.type)
        if not pointer_type:
            pointer_type = self.pointer_type

        property_name = property_name.replace('.', '_')
        template = DESERIALIZE_ELEMENT_FORMAT
        if base:
            template = DESERIALIZE_ELEMENT_BASE_FORMAT.replace('{base_property}', base.replace('*', ''))

        return get_deserialize_element_template(
            template,
            property_name,
            property_key,
            property_id,
            property_type,
            entry.has_default,
            entry.default,
            entry.status,
            pointer_type,
        )

    def get_serialize_element(self, entry: MemberVariable):
        property_name = entry.serialize_property
        property_id = entry.id
        property_key = entry.name
        property_type = replace_pointer(entry.type)
        default_value = entry.default

        assignment = '.' if self.pointer_type == 'none' else '->'
        default_argument = '' if default_value is None else f', {get_default_argument(default_value)}'
        storage_version = lookup_serialization_version(entry.version)
        conditional_serialization = storage_version != 1
        template = SERIALIZE_ELEMENT_FORMAT
        if entry.status != MemberVariableStatus.EXISTING and not conditional_serialization:
            template = "\t/* [Deleted] ({property_type}) \"{property_name}\" */\n"
        elif entry.has_default:
            template = template.replace('WriteProperty', 'WritePropertyWithDefault')
        serialization_code = template.format(
            property_name=property_name,
            property_type=property_type,
            property_id=str(property_id),
            property_key=property_key,
            property_default=default_argument,
            assignment=assignment,
        )

        if conditional_serialization:
            code = []
            if entry.status != MemberVariableStatus.EXISTING:
                # conditional delete
                code.append(f'\tif (!serializer.ShouldSerialize({storage_version})) {{')
            else:
                # conditional serialization
                code.append(f'\tif (serializer.ShouldSerialize({storage_version})) {{')
            code.append('\t' + serialization_code)

            result = '\n'.join(code) + '\t}\n'
            return result
        return serialization_code

    def generate_constructor(self, constructor_parameters: List[str]):
        parameters = ", ".join(constructor_parameters)

        if self.constructor_method is not None:
            return f'\tauto result = {self.constructor_method}({parameters});\n'
        if self.pointer_type == 'none':
            if parameters != '':
                parameters = f'({parameters})'
            return f'\t{self.return_class} result{parameters};\n'
        return f'\tauto result = duckdb::{self.pointer_type}<{self.return_class}>(new {self.return_class}({parameters}));\n'


def generate_base_class_code(base_class: SerializableClass):
    base_class_serialize = ''
    base_class_deserialize = ''

    # properties
    enum_type = ''
    for entry in base_class.members:
        if entry.serialize_property == base_class.enum_value:
            enum_type = entry.type
        base_class_serialize += base_class.get_serialize_element(entry)

        type_name = replace_pointer(entry.type)
        base_class_deserialize += base_class.get_deserialize_element(entry)
    expressions = [x for x in base_class.children.items()]
    expressions = sorted(expressions, key=lambda x: x[0])

    # set parameters
    for entry in base_class.set_parameters:
        base_class_deserialize += SET_DESERIALIZE_PARAMETER_FORMAT.format(
            property_type=entry.type, property_name=entry.name
        )

    base_class_deserialize += f'\t{base_class.pointer_type}<{base_class.name}> result;\n'
    switch_cases = ''
    for expr in expressions:
        enum_value = expr[0]
        child_data = expr[1]
        if child_data.custom_switch_code is not None:
            switch_cases += SWITCH_HEADER_FORMAT.format(
                enum_type=enum_type, enum_value=enum_value, class_deserialize=child_data.name
            )
            switch_cases += '\n'.join(
                ['\t\t' + x for x in child_data.custom_switch_code.replace('\\n', '\n').split('\n')]
            )
            switch_cases += '\n'
            continue
        switch_cases += SWITCH_STATEMENT_FORMAT.format(
            enum_type=enum_type, enum_value=enum_value, class_deserialize=child_data.name
        )

    assign_entries = []
    for entry in base_class.members:
        skip = False
        for check_entry in [entry.name, entry.serialize_property]:
            if check_entry in base_class.set_parameter_names:
                skip = True
            if check_entry == base_class.enum_value:
                skip = True
        if skip:
            continue
        assign_entries.append(entry)

    # class switch statement
    base_class_deserialize += SWITCH_CODE_FORMAT.format(
        switch_variable=base_class.enum_value, case_statements=switch_cases, base_class=base_class.name
    )

    deserialize_return = get_return_value(base_class.pointer_type, base_class.return_type)

    for entry in base_class.set_parameters:
        base_class_deserialize += UNSET_DESERIALIZE_PARAMETER_FORMAT.format(property_type=entry.type)

    for entry in assign_entries:
        if entry.status != MemberVariableStatus.EXISTING:
            continue
        move = False
        if entry.type in MOVE_LIST or is_container(entry.type) or is_pointer(entry.type):
            move = True
        if move:
            base_class_deserialize += (
                f'\tresult->{entry.deserialize_property} = std::move({entry.deserialize_property});\n'
            )
        else:
            base_class_deserialize += f'\tresult->{entry.deserialize_property} = {entry.deserialize_property};\n'
    if base_class.finalize_deserialization is not None:
        for line in base_class.finalize_deserialization:
            base_class_deserialize += "\t" + line + "\n"
    base_class_deserialize += generate_return(base_class)
    base_class_generation = ''
    serialization = ''
    if base_class.base is not None:
        serialization += BASE_SERIALIZE_FORMAT.format(base_class_name=base_class.base)
    base_class_generation += SERIALIZE_BASE_FORMAT.format(
        class_name=base_class.name, members=serialization + base_class_serialize
    )
    base_class_generation += DESERIALIZE_BASE_FORMAT.format(
        deserialize_return=deserialize_return, class_name=base_class.name, members=base_class_deserialize
    )
    return base_class_generation


def generate_class_code(class_entry: SerializableClass):
    if class_entry.custom_implementation:
        return None
    class_serialize = ''
    class_deserialize = ''

    constructor_parameters: List[str] = []
    constructor_entries = set()
    last_constructor_index = -1
    if class_entry.constructor is not None:
        for constructor_entry_ in class_entry.constructor:
            if constructor_entry_.endswith('&'):
                constructor_entry = constructor_entry_[:-1]
                is_reference = True
            else:
                constructor_entry = constructor_entry_
                is_reference = False
            constructor_entries.add(constructor_entry)
            found = False
            for entry_idx, entry in enumerate(class_entry.members):
                if entry.name == constructor_entry:
                    if entry_idx > last_constructor_index:
                        last_constructor_index = entry_idx
                    type_name = replace_pointer(entry.type)
                    entry.deserialize_property = entry.deserialize_property.replace('.', '_')
                    if requires_move(type_name) and not is_reference:
                        constructor_parameters.append(f'std::move({entry.deserialize_property})')
                    else:
                        constructor_parameters.append(entry.deserialize_property)
                    found = True
                    break

            if constructor_entry.startswith('$') or constructor_entry.startswith('?'):
                is_optional = constructor_entry.startswith('?')
                if is_optional:
                    param_type = constructor_entry.replace('?', '')
                    get_format = TRY_GET_DESERIALIZE_PARAMETER_FORMAT
                else:
                    param_type = constructor_entry.replace('$', '')
                    get_format = GET_DESERIALIZE_PARAMETER_FORMAT
                    if param_type in REFERENCE_LIST:
                        param_type += ' &'
                constructor_parameters.append(get_format.format(property_type=param_type))
                found = True

            if class_entry.base_object is not None:
                for entry in class_entry.base_object.set_parameters:
                    if entry.name == constructor_entry:
                        constructor_parameters.append(GET_DESERIALIZE_PARAMETER_FORMAT.format(property_type=entry.type))
                        found = True
                        break
            if not found:
                print(f"Constructor member \"{constructor_entry}\" was not found in members list")
                exit(1)
    elif class_entry.constructor_method is not None:
        for entry_idx, entry in enumerate(class_entry.members):
            if entry_idx > last_constructor_index:
                last_constructor_index = entry_idx
            constructor_entries.add(entry.name)
            type_name = replace_pointer(entry.type)
            entry.deserialize_property = entry.deserialize_property.replace('.', '_')
            if requires_move(type_name):
                constructor_parameters.append(f'std::move({entry.deserialize_property})')
            else:
                constructor_parameters.append(entry.deserialize_property)

    if class_entry.base is not None:
        class_serialize += BASE_SERIALIZE_FORMAT.format(base_class_name=class_entry.base)
    for entry_idx in range(last_constructor_index + 1):
        entry = class_entry.members[entry_idx]
        class_deserialize += class_entry.get_deserialize_element(entry, base=entry.base, pointer_type='unique_ptr')

    class_deserialize += class_entry.generate_constructor(constructor_parameters)
    if class_entry.members is None:
        return None
    for entry_idx, entry in enumerate(class_entry.members):
        write_property_name = entry.serialize_property
        deserialize_template_str = DESERIALIZE_ELEMENT_CLASS_FORMAT
        if entry.base:
            deserialize_template_str = DESERIALIZE_ELEMENT_CLASS_BASE_FORMAT.replace(
                '{base_property}', entry.base.replace('*', '')
            ).replace('{derived_property}', entry.type.replace('*', ''))

        class_serialize += class_entry.get_serialize_element(entry)

        type_name = replace_pointer(entry.type)
        if entry_idx > last_constructor_index:
            class_deserialize += get_deserialize_element_template(
                deserialize_template_str,
                entry.deserialize_property,
                entry.name,
                entry.id,
                type_name,
                entry.has_default,
                entry.default,
                entry.status,
                class_entry.pointer_type,
            )
        elif entry.name not in constructor_entries and entry.status == MemberVariableStatus.EXISTING:
            class_deserialize += get_deserialize_assignment(
                entry.deserialize_property, entry.type, class_entry.pointer_type
            )
        if entry.name in class_entry.set_parameter_names and entry.status == MemberVariableStatus.EXISTING:
            class_deserialize += SET_DESERIALIZE_PARAMETER_FORMAT.format(
                property_type=entry.type, property_name=entry.name
            )

    for entry in class_entry.set_parameters:
        class_deserialize += UNSET_DESERIALIZE_PARAMETER_FORMAT.format(
            property_type=entry.type, property_name=entry.name
        )
    if class_entry.finalize_deserialization is not None:
        class_deserialize += class_entry.finalize_deserialization
    if class_entry.finalize_deserialization is not None:
        for line in class_entry.finalize_deserialization:
            class_deserialize += "\t" + line + "\n"
    class_deserialize += generate_return(class_entry)
    deserialize_return = get_return_value(class_entry.pointer_type, class_entry.return_type)

    class_generation = ''
    pattern = re.compile(r'<\w+>')
    templated_type = ''

    # Check if is a templated class
    is_templated = pattern.search(class_entry.name)
    if is_templated:
        templated_type = TEMPLATED_BASE_FORMAT.format(template_name=is_templated.group()[1:-1])

    class_generation += templated_type + SERIALIZE_BASE_FORMAT.format(
        class_name=class_entry.name, members=class_serialize
    )

    class_generation += templated_type + DESERIALIZE_BASE_FORMAT.format(
        deserialize_return=deserialize_return,
        class_name=class_entry.name,
        members=class_deserialize,
    )
    return class_generation


def check_children_for_duplicate_members(node: SerializableClass, parents: list, seen_names: set, seen_ids: set):
    # Check for duplicate names
    if node.members is not None:
        for member in node.members:
            if member.name in seen_names:
                # Print the inheritance tree
                exit(
                    f"Error: Duplicate member name \"{member.name}\" in class \"{node.name}\" ({' -> '.join(map(lambda x: x.name, parents))} -> {node.name})"
                )
            seen_names.add(member.name)
            if member.id in seen_ids:
                exit(
                    f"Error: Duplicate member id \"{member.id}\" in class \"{node.name}\" ({' -> '.join(map(lambda x: x.name, parents))} -> {node.name})"
                )
            seen_ids.add(member.id)

    # Recurse
    for child in node.children.values():
        check_children_for_duplicate_members(child, parents + [node], seen_names.copy(), seen_ids.copy())


file_list = get_file_list()

for entry in file_list:
    source_path = entry['source']
    target_path = entry['target']
    with open(source_path, 'r') as f:
        try:
            json_data = json.load(f)
        except Exception as e:
            print(f"Failed to parse {source_path}: {str(e)}")
            exit(1)

    include_list = [
        'duckdb/common/serializer/serializer.hpp',
        'duckdb/common/serializer/deserializer.hpp',
    ]
    base_classes: List[SerializableClass] = []
    classes: List[SerializableClass] = []
    base_class_data: Dict[str, SerializableClass] = {}

    for entry in json_data:
        if 'includes' in entry:
            if type(entry['includes']) != type([]):
                print(f"Include list must be a list, found {type(entry['includes'])} (in {str(entry)})")
                exit(1)
            for include_entry in entry['includes']:
                if include_entry not in include_list:
                    include_list.append(include_entry)
        new_class = SerializableClass(entry)
        if new_class.is_base_class:
            # this class is a base class itself - construct the base class list
            if new_class.name in base_class_data:
                raise Exception(f"Duplicate base class \"{new_class.name}\"")
            base_class_data[new_class.name] = new_class
            base_classes.append(new_class)
        else:
            classes.append(new_class)
        if new_class.base is not None:
            # this class inherits from a base class - add the enum value
            if new_class.base not in base_class_data:
                raise Exception(f"Unknown base class \"{new_class.base}\" for entry \"{new_class.name}\"")
            base_class_object = base_class_data[new_class.base]
            new_class.inherit(base_class_object)
            for enum_entry in new_class.enum_entries:
                if enum_entry in base_class_object.children:
                    raise Exception(f"Duplicate enum entry \"{enum_entry}\"")
                base_class_object.children[enum_entry] = new_class

    # Ensure that there are no duplicate names in the inheritance tree
    for base_class in base_classes:
        if base_class.base is None:
            # Root base class, now traverse the children
            check_children_for_duplicate_members(base_class, [], set(), set())

    with open(target_path, 'w+') as f:
        include_list = ''.join([INCLUDE_FORMAT.format(filename=x) for x in include_list])
        header = HEADER.format(include_list=include_list)
        f.write(header)

        # generate the base class serialization
        for base_class in base_classes:
            base_class_generation = generate_base_class_code(base_class)
            f.write(base_class_generation)

        # generate the class serialization
        classes = sorted(classes, key=lambda x: x.name)
        for class_entry in classes:
            class_generation = generate_class_code(class_entry)
            if class_generation is None:
                continue
            f.write(class_generation)

        f.write(FOOTER)



import os
import sys
import shutil
import subprocess
from python_helpers import open_utf8
import re
import tempfile

excluded_objects = ['utf8proc_data.cpp', 'dummy_static_extension_loader.cpp']


def third_party_includes():
    includes = []
    includes += [os.path.join('third_party', 'concurrentqueue')]
    includes += [os.path.join('third_party', 'fast_float')]
    includes += [os.path.join('third_party', 'fastpforlib')]
    includes += [os.path.join('third_party', 'fmt', 'include')]
    includes += [os.path.join('third_party', 'fsst')]
    includes += [os.path.join('third_party', 'httplib')]
    includes += [os.path.join('third_party', 'hyperloglog')]
    includes += [os.path.join('third_party', 'jaro_winkler')]
    includes += [os.path.join('third_party', 'jaro_winkler', 'details')]
    includes += [os.path.join('third_party', 'libpg_query')]
    includes += [os.path.join('third_party', 'libpg_query', 'include')]
    includes += [os.path.join('third_party', 'lz4')]
    includes += [os.path.join('third_party', 'brotli', 'include')]
    includes += [os.path.join('third_party', 'brotli', 'common')]
    includes += [os.path.join('third_party', 'brotli', 'dec')]
    includes += [os.path.join('third_party', 'brotli', 'enc')]
    includes += [os.path.join('third_party', 'mbedtls', 'include')]
    includes += [os.path.join('third_party', 'mbedtls', 'library')]
    includes += [os.path.join('third_party', 'miniz')]
    includes += [os.path.join('third_party', 'pcg')]
    includes += [os.path.join('third_party', 'pdqsort')]
    includes += [os.path.join('third_party', 're2')]
    includes += [os.path.join('third_party', 'ska_sort')]
    includes += [os.path.join('third_party', 'skiplist')]
    includes += [os.path.join('third_party', 'tdigest')]
    includes += [os.path.join('third_party', 'utf8proc')]
    includes += [os.path.join('third_party', 'utf8proc', 'include')]
    includes += [os.path.join('third_party', 'vergesort')]
    includes += [os.path.join('third_party', 'yyjson', 'include')]
    includes += [os.path.join('third_party', 'zstd', 'include')]
    return includes


def third_party_sources():
    sources = []
    sources += [os.path.join('third_party', 'fmt')]
    sources += [os.path.join('third_party', 'fsst')]
    sources += [os.path.join('third_party', 'miniz')]
    sources += [os.path.join('third_party', 're2')]
    sources += [os.path.join('third_party', 'hyperloglog')]
    sources += [os.path.join('third_party', 'skiplist')]
    sources += [os.path.join('third_party', 'fastpforlib')]
    sources += [os.path.join('third_party', 'utf8proc')]
    sources += [os.path.join('third_party', 'libpg_query')]
    sources += [os.path.join('third_party', 'mbedtls')]
    sources += [os.path.join('third_party', 'yyjson')]
    sources += [os.path.join('third_party', 'zstd')]
    return sources


def file_is_lib(fname, libname):
    libextensions = ['.a', '.lib']
    libprefixes = ['', 'lib']
    for ext in libextensions:
        for prefix in libprefixes:
            potential_libname = prefix + libname + ext
            if fname == potential_libname:
                return True
    return False


def get_libraries(binary_dir, libraries, extensions):
    result_libs = []

    def find_library_recursive(search_dir, libname):
        flist = os.listdir(search_dir)
        for fname in flist:
            fpath = os.path.join(search_dir, fname)
            if os.path.isdir(fpath):
                entry = find_library_recursive(fpath, libname)
                if entry != None:
                    return entry
            elif os.path.isfile(fpath) and file_is_lib(fname, libname):
                return search_dir
        return None

    def find_library(search_dir, libname, result_libs, required=False):
        if libname == 'Threads::Threads':
            result_libs += [(None, 'pthread')]
            return
        libdir = find_library_recursive(binary_dir, libname)
        if libdir is None and required:
            raise Exception(f"Failed to locate required library {libname} in {binary_dir}")

        result_libs += [(libdir, libname)]

    duckdb_lib_name = 'duckdb_static'
    if os.name == 'nt':
        duckdb_lib_name = 'duckdb'
    find_library(os.path.join(binary_dir, 'src'), duckdb_lib_name, result_libs, True)
    for ext in extensions:
        find_library(os.path.join(binary_dir, 'extension', ext), ext + '_extension', result_libs, True)

    for libname in libraries:
        find_library(binary_dir, libname, result_libs)

    return result_libs


def includes(extensions):
    scripts_dir = os.path.dirname(os.path.abspath(__file__))
    # add includes for duckdb and extensions
    includes = []
    includes.append(os.path.join(scripts_dir, '..', 'src', 'include'))
    includes.append(os.path.join(scripts_dir, '..'))
    includes.append(os.path.join(scripts_dir, '..', 'third_party', 'utf8proc', 'include'))
    for ext in extensions:
        includes.append(os.path.join(scripts_dir, '..', 'extension', ext, 'include'))
    return includes


def include_flags(extensions):
    return ' ' + ' '.join(['-I' + x for x in includes(extensions)])


def convert_backslashes(x):
    return '/'.join(x.split(os.path.sep))


def get_relative_path(source_dir, target_file):
    source_dir = convert_backslashes(source_dir)
    target_file = convert_backslashes(target_file)

    # absolute path: try to convert
    if source_dir in target_file:
        target_file = target_file.replace(source_dir, "").lstrip('/')
    return target_file


######
# MAIN_BRANCH_VERSIONING default should be 'True' for main branch and feature branches
# MAIN_BRANCH_VERSIONING default should be 'False' for release branches
# MAIN_BRANCH_VERSIONING default value needs to keep in sync between:
# - CMakeLists.txt
# - scripts/amalgamation.py
# - scripts/package_build.py
######
MAIN_BRANCH_VERSIONING = False
if os.getenv('MAIN_BRANCH_VERSIONING') == "0":
    MAIN_BRANCH_VERSIONING = False
if os.getenv('MAIN_BRANCH_VERSIONING') == "1":
    MAIN_BRANCH_VERSIONING = True


def get_git_describe():
    override_git_describe = os.getenv('OVERRIDE_GIT_DESCRIBE') or ''
    versioning_tag_match = 'v*.*.*'
    if MAIN_BRANCH_VERSIONING:
        versioning_tag_match = 'v*.*.0'
    # empty override_git_describe, either since env was empty string or not existing
    # -> ask git (that can fail, so except in place)
    if len(override_git_describe) == 0:
        try:
            return (
                subprocess.check_output(
                    ['git', 'describe', '--tags', '--long', '--debug', '--match', versioning_tag_match]
                )
                .strip()
                .decode('utf8')
            )
        except subprocess.CalledProcessError:
            return "v0.0.0-0-gdeadbeeff"
    if len(override_git_describe.split('-')) == 3:
        return override_git_describe
    if len(override_git_describe.split('-')) == 1:
        override_git_describe += "-0"
    assert len(override_git_describe.split('-')) == 2
    try:
        return (
            override_git_describe
            + "-g"
            + subprocess.check_output(['git', 'log', '-1', '--format=%h']).strip().decode('utf8')
        )
    except subprocess.CalledProcessError:
        return override_git_describe + "-g" + "deadbeeff"


def git_commit_hash():
    if 'SETUPTOOLS_SCM_PRETEND_HASH' in os.environ:
        return os.environ['SETUPTOOLS_SCM_PRETEND_HASH']
    try:
        git_describe = get_git_describe()
        hash = git_describe.split('-')[2].lstrip('g')
        return hash
    except:
        return "deadbeeff"


def prefix_version(version):
    """Make sure the version is prefixed with 'v' to be of the form vX.Y.Z"""
    if version.startswith('v'):
        return version
    return 'v' + version


def git_dev_version():
    if 'SETUPTOOLS_SCM_PRETEND_VERSION' in os.environ:
        return prefix_version(os.environ['SETUPTOOLS_SCM_PRETEND_VERSION'])
    try:
        long_version = get_git_describe()
        version_splits = long_version.split('-')[0].lstrip('v').split('.')
        dev_version = long_version.split('-')[1]
        if int(dev_version) == 0:
            # directly on a tag: emit the regular version
            return "v" + '.'.join(version_splits)
        else:
            # not on a tag: increment the version by one and add a -devX suffix
            # this needs to keep in sync with changes to CMakeLists.txt
            if MAIN_BRANCH_VERSIONING == True:
                # increment minor version
                version_splits[1] = str(int(version_splits[1]) + 1)
            else:
                # increment patch version
                version_splits[2] = str(int(version_splits[2]) + 1)
            return "v" + '.'.join(version_splits) + "-dev" + dev_version
    except:
        return "v0.0.0"


def include_package(pkg_name, pkg_dir, include_files, include_list, source_list):
    import amalgamation

    original_path = sys.path
    # append the directory
    sys.path.append(pkg_dir)
    ext_pkg = __import__(pkg_name + '_config')

    ext_include_dirs = ext_pkg.include_directories
    ext_source_files = ext_pkg.source_files

    include_files += amalgamation.list_includes_files(ext_include_dirs)
    include_list += ext_include_dirs
    source_list += ext_source_files

    sys.path = original_path


def get_extension_linked_define(extension):
    return f'DUCKDB_EXTENSION_{extension.upper()}_LINKED'


def build_package(
    target_dir,
    extensions,
    linenumbers=False,
    unity_count=32,
    folder_name='duckdb',
    short_paths=False,
    default_linked_extensions=None,
):
    if not os.path.isdir(target_dir):
        os.mkdir(target_dir)

    extensions = list(extensions)
    # Keep existing package_build behavior by default: all packaged extensions are linked.
    # Callers that package a superset can pass default_linked_extensions to emit a loader
    # that is controlled by DUCKDB_EXTENSION_<NAME>_LINKED compile definitions instead.
    if default_linked_extensions is None:
        default_linked_extensions = extensions
    default_linked_extensions = set(default_linked_extensions)
    packaged_extensions = set(extensions)
    unpackaged_linked_extensions = default_linked_extensions - packaged_extensions
    if unpackaged_linked_extensions:
        raise ValueError(
            "default_linked_extensions must be a subset of extensions: {}".format(
                ', '.join(sorted(unpackaged_linked_extensions))
            )
        )

    scripts_dir = os.path.dirname(os.path.abspath(__file__))
    sys.path.append(scripts_dir)
    import amalgamation

    prev_wd = os.getcwd()
    os.chdir(os.path.join(scripts_dir, '..'))

    # obtain the list of source files from the amalgamation
    source_list = amalgamation.list_sources()
    include_list = amalgamation.list_include_dirs()
    include_files = amalgamation.list_includes()

    def copy_file(src, target_dir):
        # get the path
        full_path = src.split(os.path.sep)
        current_path = target_dir
        for i in range(len(full_path) - 1):
            current_path = os.path.join(current_path, full_path[i])
            if not os.path.isdir(current_path):
                os.mkdir(current_path)
        target_name = full_path[-1]
        target_file = os.path.join(current_path, target_name)
        amalgamation.copy_if_different(src, target_file)

    # include the main extension helper
    include_files += [os.path.join('src', 'include', 'duckdb', 'main', 'extension_helper.hpp')]
    # include the separate extensions
    ext_loader_body = ''
    ext_loader_defines = ''
    ext_headers = ''
    ext_name_vector_initializer = ''
    for ext in extensions:
        ext_path = os.path.join(scripts_dir, '..', 'extension', ext)
        include_package(ext, ext_path, include_files, include_list, source_list)

        ext_linked_define = get_extension_linked_define(ext)
        ext_linked_default = 1 if ext in default_linked_extensions else 0

        ext_loader_defines += (
            f"#ifndef {ext_linked_define}\n" f"#define {ext_linked_define} {ext_linked_default}\n" "#endif\n\n"
        )

        ext_headers += f'#if {ext_linked_define}\n#include "{ext}_extension.hpp"\n#endif\n'

        # handle generated_extension_loader
        # this - beautifully - approximates code in extension/CMakeLists.txt
        ext_name_camelcase = ext.replace('_', ' ').title().replace(' ', '')

        ext_loader_body += (
            f"#if {ext_linked_define}\n"
            f"    if (extension==\"{ext}\") {{\n"
            f"        db.LoadStaticExtension<{ext_name_camelcase}Extension>();\n"
            "        return ExtensionLoadResult::LOADED_EXTENSION;\n"
            "    }\n"
            "#endif\n"
        )

        ext_name_vector_initializer += f"\n#if {ext_linked_define}\n" f"        \"{ext}\",\n" "#endif"

    loader_code = open(os.path.join('extension', 'generated_extension_loader.cpp.in'), 'rb').read().decode('utf8')
    loader_code = (
        loader_code.replace('${EXT_LOADER_BODY}', ext_loader_body)
        .replace('${EXT_NAME_VECTOR_INITIALIZER}', ext_name_vector_initializer)
        .replace('${EXT_TEST_PATH_INITIALIZER}', '')
        .replace('CMake', 'package_build.py')
    )

    loader_code = ext_loader_defines + ext_headers + loader_code

    loader_name = 'generated_extension_loader_package_build.cpp'
    f = open(loader_name, 'wb')
    f.write(loader_code.encode('utf8'))
    f.close()

    source_list += [loader_name]

    for src in source_list:
        copy_file(src, target_dir)

    for inc in include_files:
        copy_file(inc, target_dir)

    # handle pragma_version.cpp: paste #define DUCKDB_SOURCE_ID and DUCKDB_VERSION there
    curdir = os.getcwd()
    os.chdir(os.path.join(scripts_dir, '..'))
    githash = git_commit_hash()
    dev_version = git_dev_version()
    dev_v_parts = dev_version.lstrip('v').split('.')
    os.chdir(curdir)
    # open the file and read the current contents
    fpath = os.path.join(target_dir, 'src', 'function', 'table', 'version', 'pragma_version.cpp')
    with open_utf8(fpath, 'r') as f:
        text = f.read()
    # now add the DUCKDB_SOURCE_ID define, if it is not there already
    found_hash = False
    found_dev = False
    found_major = False
    found_minor = False
    found_patch = False
    lines = text.split('\n')
    for i in range(len(lines)):
        if '#define DUCKDB_SOURCE_ID ' in lines[i]:
            lines[i] = '#define DUCKDB_SOURCE_ID "{}"'.format(githash)
            found_hash = True
        if '#define DUCKDB_VERSION ' in lines[i]:
            lines[i] = '#define DUCKDB_VERSION "{}"'.format(dev_version)
            found_dev = True
        if '#define DUCKDB_MAJOR_VERSION ' in lines[i]:
            lines[i] = '#define DUCKDB_MAJOR_VERSION {}'.format(int(dev_v_parts[0]))
            found_major = True
        if '#define DUCKDB_MINOR_VERSION ' in lines[i]:
            lines[i] = '#define DUCKDB_MINOR_VERSION {}'.format(int(dev_v_parts[1]))
            found_minor = True
        if '#define DUCKDB_PATCH_VERSION ' in lines[i]:
            lines[i] = '#define DUCKDB_PATCH_VERSION "{}"'.format(dev_v_parts[2])
            found_patch = True
    if not found_hash:
        lines = ['#ifndef DUCKDB_SOURCE_ID', '#define DUCKDB_SOURCE_ID "{}"'.format(githash), '#endif'] + lines
    if not found_dev:
        lines = ['#ifndef DUCKDB_VERSION', '#define DUCKDB_VERSION "{}"'.format(dev_version), '#endif'] + lines
    if not found_major:
        lines = [
            '#ifndef DUCKDB_MAJOR_VERSION',
            '#define DUCKDB_MAJOR_VERSION {}'.format(int(dev_v_parts[0])),
            '#endif',
        ] + lines
    if not found_minor:
        lines = [
            '#ifndef DUCKDB_MINOR_VERSION',
            '#define DUCKDB_MINOR_VERSION {}'.format(int(dev_v_parts[1])),
            '#endif',
        ] + lines
    if not found_patch:
        lines = [
            '#ifndef DUCKDB_PATCH_VERSION',
            '#define DUCKDB_PATCH_VERSION "{}"'.format(dev_v_parts[2]),
            '#endif',
        ] + lines
    text = '\n'.join(lines)
    with open_utf8(fpath, 'w+') as f:
        f.write(text)

    def file_is_excluded(fname):
        for entry in excluded_objects:
            if entry in fname:
                return True
        return False

    def generate_unity_build(entries, unity_name, linenumbers):
        ub_file = os.path.join(target_dir, unity_name)
        with open_utf8(ub_file, 'w+') as f:
            for entry in entries:
                if linenumbers:
                    f.write('#line 0 "{}"\n'.format(convert_backslashes(entry)))
                f.write('#include "{}"\n\n'.format(convert_backslashes(entry)))
        return ub_file

    def generate_unity_builds(source_list, nsplits, linenumbers):
        files_per_directory = {}
        for source in source_list:
            dirname = os.path.dirname(source)
            if dirname not in files_per_directory:
                files_per_directory[dirname] = []
            files_per_directory[dirname].append(source)

        new_source_files = []
        for dirname in files_per_directory.keys():
            current_files = files_per_directory[dirname]
            cmake_file = os.path.join(dirname, 'CMakeLists.txt')
            unity_build = False
            if os.path.isfile(cmake_file) and len(current_files) > 1:
                with open(cmake_file, 'r') as f:
                    text = f.read()
                    if 'add_library_unity' in text:
                        unity_build = True
                        # re-order the files in the unity build so that they follow the same order as the CMake
                        scores = {}
                        filenames = [x[0] for x in re.findall('([a-zA-Z0-9_]+[.](cpp|cc|c|cxx))', text)]
                        score = 0
                        for filename in filenames:
                            scores[filename] = score
                            score += 1
                        current_files.sort(
                            key=lambda x: scores[os.path.basename(x)] if os.path.basename(x) in scores else 99999
                        )
            if not unity_build:
                if short_paths:
                    # replace source files with "__"
                    for file in current_files:
                        unity_filename = os.path.basename(file)
                        new_source_files.append(generate_unity_build([file], unity_filename, linenumbers))
                else:
                    # directly use the source files
                    new_source_files += [os.path.join(folder_name, file) for file in current_files]
            else:
                unity_base = dirname.replace(os.path.sep, '_')
                unity_name = f'ub_{unity_base}.cpp'
                new_source_files.append(generate_unity_build(current_files, unity_name, linenumbers))
        return new_source_files

    original_sources = source_list
    source_list = generate_unity_builds(source_list, unity_count, linenumbers)

    os.chdir(prev_wd)
    return (
        [convert_backslashes(x) for x in source_list if not file_is_excluded(x)],
        [convert_backslashes(x) for x in include_list],
        [convert_backslashes(x) for x in original_sources],
    )



# fmt: off

import pytest
import subprocess
import sys
from typing import List
from conftest import ShellTest
import os

def test_invalid_explain(shell):
    test = (
        ShellTest(shell)
        .statement("EXPLAIN SELECT 'any_string' IN ?;")
    )
    result = test.run()



import subprocess
import os
import argparse
import re

parser = argparse.ArgumentParser(description='Publish a Julia release.')
parser.add_argument(
    '--yggdrassil-fork',
    dest='yggdrassil',
    action='store',
    help='Fork of the Julia Yggdrassil repository (https://github.com/JuliaPackaging/Yggdrasil)',
    default='/Users/myth/Programs/Yggdrasil',
)
args = parser.parse_args()

if not os.path.isfile(os.path.join('tools', 'juliapkg', 'release.py')):
    print('This script must be run from the root DuckDB directory (i.e. `python3 tools/juliapkg/release.py`)')
    exit(1)


def run_syscall(syscall, ignore_failure=False):
    res = os.system(syscall)
    if ignore_failure:
        return
    if res != 0:
        print(f'Failed to execute {syscall}: got exit code {str(res)}')
        exit(1)


# helper script to generate a julia release
duckdb_path = os.getcwd()

# fetch the latest tags
os.system('git fetch upstream --tags')

proc = subprocess.Popen(['git', 'show-ref', '--tags'], stdout=subprocess.PIPE)
tags = [x for x in proc.stdout.read().decode('utf8').split('\n') if len(x) > 0 and 'master-builds' not in x]


def extract_tag(x):
    keys = x.split('refs/tags/')[1].lstrip('v').split('.')
    return int(keys[0]) * 10000000 + int(keys[1]) * 10000 + int(keys[2])


tags.sort(key=extract_tag)

# latest tag
splits = tags[-1].split(' ')
hash = splits[0]
tag = splits[1].replace('refs/tags/', '')
if tag[0] != 'v':
    print(f"Tag {tag} does not start with a v?")
    exit(1)

print(f'Creating a Julia release from the latest tag {tag} with commit hash {hash}')

print('> Creating a PR to the Yggdrassil repository (https://github.com/JuliaPackaging/Yggdrasil)')

os.chdir(args.yggdrassil)
run_syscall('git checkout master')
run_syscall('git pull upstream master')
run_syscall(f'git branch -D {tag}', True)
run_syscall(f'git checkout -b {tag}')
tarball_build = os.path.join('D', 'DuckDB', 'build_tarballs.jl')
with open(tarball_build, 'r') as f:
    text = f.read()

text = re.sub('\nversion = v["][0-9.]+["]\n', f'\nversion = v"{tag[1:]}"\n', text)
text = re.sub(
    'GitSource[(]["]https[:][/][/]github[.]com[/]duckdb[/]duckdb[.]git["][,] ["][a-zA-Z0-9]+["][)]',
    f'GitSource("https://github.com/duckdb/duckdb.git", "{hash}")',
    text,
)

with open(tarball_build, 'w+') as f:
    f.write(text)

run_syscall(f'git add {tarball_build}')
run_syscall(f'git commit -m "[DuckDB] Bump to {tag}"')
run_syscall(f'git push --set-upstream origin {tag}')
run_syscall(
    f'gh pr create --title "[DuckDB] Bump to {tag}" --repo "https://github.com/JuliaPackaging/Yggdrasil" --body ""'
)

print('PR has been created.\n')
print(f'Next up we need to bump the version and DuckDB_jll version to {tag} in `tools/juliapkg/Project.toml`')
print('This is not yet automated.')
print(
    '> After that PR is merged - we need to post a comment containing the text `@JuliaRegistrator register subdir=tools/juliapkg`'
)
print('> For example, see https://github.com/duckdb/duckdb/commit/0f0461113f3341135471805c9928c4d71d1f5874')



import argparse
import logging
import os
import pathlib
import re
from types import NoneType
from typing import Dict, List, NotRequired, TypedDict, Union

from generate_c_api import (
    EXT_API_DEFINITION_PATTERN,
    get_extension_api_version,
    parse_capi_function_definitions,
    parse_ext_api_definitions,
)


class FunctionDefParam(TypedDict):
    type: str
    name: str


class FunctionDefComment(TypedDict):
    description: str
    param_comments: dict[str, str]
    return_value: str


class FunctionDef(TypedDict):
    name: str
    group: str
    deprecated: bool
    group_deprecated: bool
    return_type: str
    params: list[FunctionDefParam]
    comment: FunctionDefComment


class FunctionGroup(TypedDict):
    group: str
    deprecated: bool
    entries: list[FunctionDef]


class DuckDBApiInfo(TypedDict):
    version: str
    commit: NotRequired[str]


def parse_c_type(type_str: str, type: list[str] = []):
    """Parses simple C types (no function pointer or array types) and returns a list of the type components.

    Args:
        type_str: A C type string to parse, e.g.: "const char* const"
        type: List to track components, used for recursion. Defaults to [].

    Returns:
        list: A list of the type components, e.g.: "const char* const" -> ["Const Ptr", "const char"]
    """
    type_str = type_str.strip()
    ptr_pattern = r"^(.*)\*(\s*const\s*)?$"

    if (m1 := re.match(ptr_pattern, type_str)) is not None:
        before_ptr = m1.group(1)
        is_const = bool(m1.group(2))
        type.append("Const Ptr" if is_const else "Ptr")
        return parse_c_type(before_ptr, type)

    type.append(type_str)
    return type


JULIA_RESERVED_KEYWORDS = {
    "function",
    "if",
    "else",
    "while",
    "for",
    "try",
    "catch",
    "finally",
    "return",
    "break",
    "continue",
    "end",
    "begin",
    "quote",
    "let",
    "local",
    "global",
    "const",
    "do",
    "struct",
    "mutable",
    "abstract",
    "type",
    "module",
    "using",
    "import",
    "export",
    "public",
}

JULIA_BASE_TYPE_MAP = {
    # Julia Standard Types
    "char": "Char",
    "int": "Int",
    "int8_t": "Int8",
    "int16_t": "Int16",
    "int32_t": "Int32",
    "int64_t": "Int64",
    "uint8_t": "UInt8",
    "uint16_t": "UInt16",
    "uint32_t": "UInt32",
    "uint64_t": "UInt64",
    "double": "Float64",
    "float": "Float32",
    "bool": "Bool",
    "void": "Cvoid",
    "size_t": "Csize_t",
    # DuckDB specific types
    "idx_t": "idx_t",
    "duckdb_type": "DUCKDB_TYPE",
    "duckdb_string_t": "duckdb_string_t",  # INLINE prefix with pointer string type
    "duckdb_string": "duckdb_string",  # Pointer + size type
    "duckdb_table_function": "duckdb_table_function",  # actually struct pointer
    "duckdb_table_function_t": "duckdb_table_function_ptr",  # function pointer type
    "duckdb_cast_function": "duckdb_cast_function",  # actually struct pointer
    "duckdb_cast_function_t": "duckdb_cast_function_ptr",  # function pointer type
}


# TODO this the original order of the functions in `api.jl` and is only used to keep the PR review small
JULIA_API_ORIGINAL_ORDER = [
    "duckdb_open",
    "duckdb_open_ext",
    "duckdb_close",
    "duckdb_connect",
    "duckdb_disconnect",
    "duckdb_create_config",
    "duckdb_config_count",
    "duckdb_get_config_flag",
    "duckdb_set_config",
    "duckdb_destroy_config",
    "duckdb_query",
    "duckdb_destroy_result",
    "duckdb_column_name",
    "duckdb_column_type",
    "duckdb_column_logical_type",
    "duckdb_column_count",
    "duckdb_row_count",
    "duckdb_rows_changed",
    "duckdb_column_data",
    "duckdb_nullmask_data",
    "duckdb_result_error",
    "duckdb_result_get_chunk",
    "duckdb_result_is_streaming",
    "duckdb_stream_fetch_chunk",
    "duckdb_result_chunk_count",
    "duckdb_value_boolean",
    "duckdb_value_int8",
    "duckdb_value_int16",
    "duckdb_value_int32",
    "duckdb_value_int64",
    "duckdb_value_hugeint",
    "duckdb_value_uhugeint",
    "duckdb_value_uint8",
    "duckdb_value_uint16",
    "duckdb_value_uint32",
    "duckdb_value_uint64",
    "duckdb_value_float",
    "duckdb_value_double",
    "duckdb_value_date",
    "duckdb_value_time",
    "duckdb_value_timestamp",
    "duckdb_value_interval",
    "duckdb_value_varchar",
    "duckdb_value_varchar_internal",
    "duckdb_value_is_null",
    "duckdb_malloc",
    "duckdb_free",
    "duckdb_vector_size",
    "duckdb_from_time_tz",
    "duckdb_prepare",
    "duckdb_destroy_prepare",
    "duckdb_prepare_error",
    "duckdb_nparams",
    "duckdb_param_type",
    "duckdb_bind_boolean",
    "duckdb_bind_int8",
    "duckdb_bind_int16",
    "duckdb_bind_int32",
    "duckdb_bind_int64",
    "duckdb_bind_hugeint",
    "duckdb_bind_uhugeint",
    "duckdb_bind_uint8",
    "duckdb_bind_uint16",
    "duckdb_bind_uint32",
    "duckdb_bind_uint64",
    "duckdb_bind_float",
    "duckdb_bind_double",
    "duckdb_bind_date",
    "duckdb_bind_time",
    "duckdb_bind_timestamp",
    "duckdb_bind_interval",
    "duckdb_bind_varchar",
    "duckdb_bind_varchar_length",
    "duckdb_bind_blob",
    "duckdb_bind_null",
    "duckdb_execute_prepared",
    "duckdb_pending_prepared",
    "duckdb_pending_prepared_streaming",
    "duckdb_pending_execute_check_state",
    "duckdb_destroy_pending",
    "duckdb_pending_error",
    "duckdb_pending_execute_task",
    "duckdb_execute_pending",
    "duckdb_pending_execution_is_finished",
    "duckdb_destroy_value",
    "duckdb_create_varchar",
    "duckdb_create_varchar_length",
    "duckdb_create_int64",
    "duckdb_get_varchar",
    "duckdb_get_int64",
    "duckdb_create_logical_type",
    "duckdb_create_decimal_type",
    "duckdb_get_type_id",
    "duckdb_decimal_width",
    "duckdb_decimal_scale",
    "duckdb_decimal_internal_type",
    "duckdb_enum_internal_type",
    "duckdb_enum_dictionary_size",
    "duckdb_enum_dictionary_value",
    "duckdb_list_type_child_type",
    "duckdb_struct_type_child_count",
    "duckdb_union_type_member_count",
    "duckdb_struct_type_child_name",
    "duckdb_union_type_member_name",
    "duckdb_struct_type_child_type",
    "duckdb_union_type_member_type",
    "duckdb_destroy_logical_type",
    "duckdb_create_data_chunk",
    "duckdb_destroy_data_chunk",
    "duckdb_data_chunk_reset",
    "duckdb_data_chunk_get_column_count",
    "duckdb_data_chunk_get_size",
    "duckdb_data_chunk_set_size",
    "duckdb_data_chunk_get_vector",
    "duckdb_vector_get_column_type",
    "duckdb_vector_get_data",
    "duckdb_vector_get_validity",
    "duckdb_vector_ensure_validity_writable",
    "duckdb_list_vector_get_child",
    "duckdb_list_vector_get_size",
    "duckdb_struct_vector_get_child",
    "duckdb_union_vector_get_member",
    "duckdb_vector_assign_string_element",
    "duckdb_vector_assign_string_element_len",
    "duckdb_create_table_function",
    "duckdb_destroy_table_function",
    "duckdb_table_function_set_name",
    "duckdb_table_function_add_parameter",
    "duckdb_table_function_set_extra_info",
    "duckdb_table_function_set_bind",
    "duckdb_table_function_set_init",
    "duckdb_table_function_set_local_init",
    "duckdb_table_function_set_function",
    "duckdb_table_function_supports_projection_pushdown",
    "duckdb_register_table_function",
    "duckdb_bind_get_extra_info",
    "duckdb_bind_add_result_column",
    "duckdb_bind_get_parameter_count",
    "duckdb_bind_get_parameter",
    "duckdb_bind_set_bind_data",
    "duckdb_bind_set_cardinality",
    "duckdb_bind_set_error",
    "duckdb_init_get_extra_info",
    "duckdb_init_get_bind_data",
    "duckdb_init_set_init_data",
    "duckdb_init_get_column_count",
    "duckdb_init_get_column_index",
    "duckdb_init_set_max_threads",
    "duckdb_init_set_error",
    "duckdb_function_get_extra_info",
    "duckdb_function_get_bind_data",
    "duckdb_function_get_init_data",
    "duckdb_function_get_local_init_data",
    "duckdb_function_set_error",
    "duckdb_add_replacement_scan",
    "duckdb_replacement_scan_set_function_name",
    "duckdb_replacement_scan_add_parameter",
    "duckdb_replacement_scan_set_error",
    "duckdb_appender_create",
    "duckdb_appender_error",
    "duckdb_appender_flush",
    "duckdb_appender_close",
    "duckdb_appender_destroy",
    "duckdb_appender_begin_row",
    "duckdb_appender_end_row",
    "duckdb_append_bool",
    "duckdb_append_int8",
    "duckdb_append_int16",
    "duckdb_append_int32",
    "duckdb_append_int64",
    "duckdb_append_hugeint",
    "duckdb_append_uhugeint",
    "duckdb_append_uint8",
    "duckdb_append_uint16",
    "duckdb_append_uint32",
    "duckdb_append_uint64",
    "duckdb_append_float",
    "duckdb_append_double",
    "duckdb_append_date",
    "duckdb_append_time",
    "duckdb_append_timestamp",
    "duckdb_append_interval",
    "duckdb_append_varchar",
    "duckdb_append_varchar_length",
    "duckdb_append_blob",
    "duckdb_append_null",
    "duckdb_execute_tasks",
    "duckdb_create_task_state",
    "duckdb_execute_tasks_state",
    "duckdb_execute_n_tasks_state",
    "duckdb_finish_execution",
    "duckdb_task_state_is_finished",
    "duckdb_destroy_task_state",
    "duckdb_execution_is_finished",
    "duckdb_create_scalar_function",
    "duckdb_destroy_scalar_function",
    "duckdb_scalar_function_set_name",
    "duckdb_scalar_function_add_parameter",
    "duckdb_scalar_function_set_return_type",
    "duckdb_scalar_function_set_function",
    "duckdb_register_scalar_function",
]


class JuliaApiTarget:
    indent: int = 0
    linesep: str = os.linesep
    type_maps: dict[str, str] = {}  # C to Julia
    inverse_type_maps: dict[str, list[str]] = {}  # Julia to C
    deprecated_functions: list[str] = []
    type_map: dict[str, str]

    # Functions to skip
    skipped_functions = set()
    skip_deprecated_functions = False

    # Explicit function order
    manual_order: Union[List[str], NoneType] = None

    overwrite_function_signatures = {}

    # Functions that use indices either as ARG or RETURN and should be converted to 1-based indexing
    auto_1base_index: bool
    auto_1base_index_return_functions = set()
    auto_1base_index_ignore_functions = set()

    def __init__(
        self,
        file,
        indent=0,
        auto_1base_index=True,
        auto_1base_index_return_functions=set(),
        auto_1base_index_ignore_functions=set(),
        skipped_functions=set(),
        skip_deprecated_functions=False,
        type_map={},
        overwrite_function_signatures={},
    ):
        # check if file is a string or a file object
        if isinstance(file, str) or isinstance(file, pathlib.Path):
            self.filename = pathlib.Path(file)
        else:
            raise ValueError("file must be a string or a path object")
        self.indent = indent
        self.auto_1base_index = auto_1base_index
        self.auto_1base_index_return_functions = auto_1base_index_return_functions
        self.auto_1base_index_ignore_functions = auto_1base_index_ignore_functions
        self.linesep = os.linesep
        self.type_map = type_map
        self.skipped_functions = skipped_functions
        self.skip_deprecated_functions = skip_deprecated_functions
        self.overwrite_function_signatures = overwrite_function_signatures
        super().__init__()

    def __enter__(self):
        self.file = open(self.filename, "w")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.file.close()

    def write_empty_line(self, n=1) -> None:
        """Writes an empty line to the output file."""
        for i in range(n):
            self.file.write(self.linesep)

    def _get_casted_type(self, type_str: str, is_return_arg=False, auto_remove_t_suffix=True):
        type_str = type_str.strip()
        type_definition = parse_c_type(type_str, [])

        def reduce_type(type_list: list[str]):
            if len(type_list) == 0:
                return ""

            t = type_list[0]
            if len(type_list) == 1:
                is_const = False  #  Track that the type is const, even though we cannot use it in Julia
                if t.startswith("const "):
                    t, is_const = t.removeprefix("const "), True

                if t in self.type_map:
                    return self.type_map[t]
                else:
                    if auto_remove_t_suffix and t.endswith("_t"):
                        t = t.removesuffix("_t")
                    if " " in t:
                        raise (ValueError(f"Unknown type: {t}"))
                    return t

            # Handle Pointer types
            if t not in ("Ptr", "Const Ptr"):
                raise ValueError(f"Unexpected non-pointer type: {t}")

            if len(type_list) >= 2 and type_list[1].strip() in (
                "char",
                "const char",
            ):
                return "Cstring"
            else:
                if is_return_arg:
                    # Use Ptr for return types, because they are not tracked by the Julia GC
                    return "Ptr{" + reduce_type(type_list[1:]) + "}"
                else:
                    # Prefer Ref over Ptr for arguments
                    return "Ref{" + reduce_type(type_list[1:]) + "}"

        return reduce_type(type_definition)

    def _is_index_argument(self, name: str, function_obj: FunctionDef):
        # Check if the argument is (likely) an index
        if name not in (
            "index",
            "idx",
            "i",
            "row",
            "col",
            "column",
            "col_idx",
            "column_idx",
            "column_index",
            "row_idx",
            "row_index",
            "chunk_index",
            # "param_idx", # TODO creates errors in bind_param
        ):
            return False

        x = None
        for param in function_obj["params"]:
            if param["name"] == name:
                x = param
                break

        arg_type = self._get_casted_type(x["type"])
        if arg_type not in (
            "Int",
            "Int64",
            "UInt",
            "UInt64",
            "idx_t",
            "idx" "Int32",
            "UInt32",
            "Csize_t",
        ):
            return False

        return True

    def get_argument_names_and_types(self, function_obj: FunctionDef):
        def _get_arg_name(name: str):
            if name in JULIA_RESERVED_KEYWORDS:
                return f"_{name}"
            return name

        arg_names = [_get_arg_name(param["name"]) for param in function_obj["params"]]

        if function_obj["name"] in self.overwrite_function_signatures:
            return_type, arg_types = self.overwrite_function_signatures[function_obj["name"]]
            return arg_names, arg_types

        arg_types = [self._get_casted_type(param["type"]) for param in function_obj["params"]]
        return arg_names, arg_types

    def is_index1_function(self, function_obj: FunctionDef):
        fname = function_obj["name"]

        if not self.auto_1base_index:
            return [False for param in function_obj["params"]], False

        if fname in self.auto_1base_index_ignore_functions:
            return [False for param in function_obj["params"]], False

        is_index1_return = fname in self.auto_1base_index_return_functions
        is_index1_arg = [self._is_index_argument(param["name"], function_obj) for param in function_obj["params"]]
        return is_index1_arg, is_index1_return

    def _write_function_docstring(self, function_obj: FunctionDef):
        r"""_create_function_docstring


        Example:
        ```julia
        \"\"\"
            duckdb_get_int64(value)

        Obtains an int64 of the given value.

        # Arguments
        - `value`: The value

        Returns: The int64 value, or 0 if no conversion is possible
        \"\"\"
        ```

        Args:
            function_obj: _description_
        """

        description = function_obj.get("comment", {}).get("description", "").strip()
        description = description.replace('"', '\\"')  # escape double quotes

        index1_args, index1_return = self.is_index1_function(function_obj)

        # Arguments
        arg_names, arg_types = self.get_argument_names_and_types(function_obj)

        arg_comments = []
        for ix, (name, param, t, is_index1) in enumerate(
            zip(arg_names, function_obj["params"], arg_types, index1_args)
        ):
            param_comment = function_obj.get("comment", {}).get("param_comments", {}).get(param["name"], "")
            if is_index1:
                parts = [f"`{name}`:", f"`{t}`", "(1-based index)", param_comment]
            else:
                parts = [f"`{name}`:", f"`{t}`", param_comment]
            arg_comments.append(" ".join(parts))

        arg_names_s = ", ".join(arg_names)

        # Return Values
        return_type = self._get_casted_type(function_obj["return_type"], is_return_arg=True)
        if return_type == "Cvoid":
            return_type = "Nothing"  # Cvoid is equivalent to Nothing in Julia
        return_comments = [
            f"`{return_type}`",
            function_obj.get("comment", {}).get("return_value", ""),
        ]
        if index1_return:
            return_comments.append("(1-based index)")
        return_value_comment = " ".join(return_comments)

        self.file.write(f"{'    ' * self.indent}\"\"\"\n")
        self.file.write(f"{'    ' * self.indent}    {function_obj['name']}({arg_names_s})\n")
        self.file.write(f"{'    ' * self.indent}\n")
        self.file.write(f"{'    ' * self.indent}{description}\n")
        self.file.write(f"{'    ' * self.indent}\n")
        self.file.write(f"{'    ' * self.indent}# Arguments\n")
        for i, arg_name in enumerate(arg_names):
            self.file.write(f"{'    ' * self.indent}- {arg_comments[i]}\n")
        self.file.write(f"{'    ' * self.indent}\n")
        self.file.write(f"{'    ' * self.indent}Returns: {return_value_comment}\n")
        self.file.write(f"{'    ' * self.indent}\"\"\"\n")

    def _get_depwarning_message(self, function_obj: FunctionDef):
        description = function_obj.get("comment", {}).get("description", "")
        if not description.startswith("**DEPRECATION NOTICE**:"):
            description = f"**DEPRECATION NOTICE**: {description}"

        # Only use the first line of the description
        notice = description.split("\n")[0]
        notice = notice.replace("\n", " ").replace('"', '\\"').strip()
        return notice

    def _write_function_depwarn(self, function_obj: FunctionDef, indent: int = 0):
        """
        Writes a deprecation warning for a function.

        Example:
        ```julia
            Base.depwarn(
            "The `G` type parameter will be deprecated in a future release. " *
            "Please use `MyType(args...)` instead of `MyType{$G}(args...)`.",
            :MyType,
            )
        ```
        """
        indent = self.indent + indent  # total indent

        notice = self._get_depwarning_message(function_obj)

        self.file.write(f"{'    ' * indent}Base.depwarn(\n")
        self.file.write(f"{'    ' * indent}  \"{notice}\",\n")
        self.file.write(f"{'    ' * indent}    :{function_obj['name']},\n")
        self.file.write(f"{'    ' * indent})\n")

    def _list_to_julia_tuple(self, lst):
        if len(lst) == 0:
            return "()"
        elif len(lst) == 1:
            return f"({lst[0]},)"
        else:
            return f"({', '.join(lst)})"

    def _write_function_definition(self, function_obj: FunctionDef):
        fname = function_obj["name"]
        index1_args, index1_return = self.is_index1_function(function_obj)

        arg_names, arg_types = self.get_argument_names_and_types(function_obj)
        arg_types_tuple = self._list_to_julia_tuple(arg_types)
        arg_names_definition = ", ".join(arg_names)

        arg_names_call = []
        for arg_name, is_index1 in zip(arg_names, index1_args):
            if is_index1:
                arg_names_call.append(f"{arg_name} - 1")
            else:
                arg_names_call.append(arg_name)
        arg_names_call = ", ".join(arg_names_call)

        return_type = self._get_casted_type(function_obj["return_type"], is_return_arg=True)

        self.file.write(f"{'    ' * self.indent}function {fname}({arg_names_definition})\n")

        if function_obj.get("group_deprecated", False) or function_obj.get("deprecated", False):
            self._write_function_depwarn(function_obj, indent=1)

        self.file.write(
            f"{'    ' * self.indent}    return ccall((:{fname}, libduckdb), {return_type}, {arg_types_tuple}, {arg_names_call}){' + 1' if index1_return else ''}\n"
        )
        self.file.write(f"{'    ' * self.indent}end\n")

    def write_function(self, function_obj: FunctionDef):
        if function_obj["name"] in self.skipped_functions:
            return

        if function_obj.get("group_deprecated", False) or function_obj.get("deprecated", False):
            self.deprecated_functions.append(function_obj["name"])

        self._write_function_docstring(function_obj)
        self._write_function_definition(function_obj)

    def write_footer(self):
        self.write_empty_line(n=1)
        s = """
# !!!!!!!!!!!!
# WARNING: this file is autogenerated by scripts/generate_c_api_julia.py, manual changes will be overwritten
# !!!!!!!!!!!!
"""
        self.file.write(s)
        self.write_empty_line()

    def write_header(self, version=""):
        s = """
###############################################################################
# 
# DuckDB Julia API
# 
# !!!!!!!!!!!!
# WARNING: this file is autogenerated by scripts/generate_c_api_julia.py, manual changes will be overwritten
# !!!!!!!!!!!!
#
###############################################################################

using Base.Libc

if "JULIA_DUCKDB_LIBRARY" in keys(ENV)
    libduckdb = ENV["JULIA_DUCKDB_LIBRARY"]
else
    using DuckDB_jll
end
"""
        if version[0] == "v":
            # remove the v prefix and use Julia Version String
            version = version[1:]

        self.file.write(s)
        self.file.write("\n")
        self.file.write(f'DUCKDB_API_VERSION = v"{version}"\n')
        self.file.write("\n")

    def write_functions(
        self,
        version,
        function_groups: List[FunctionGroup],
        function_map: Dict[str, FunctionDef],
    ):
        self._analyze_types(function_groups)  # Create the julia type map
        self.write_header(version)
        self.write_empty_line()
        if self.manual_order is not None:
            current_group = None
            for f in self.manual_order:
                if f not in function_map:
                    print(f"WARNING: Function {f} not found in function_map")
                    continue

                if current_group != function_map[f]["group"]:
                    current_group = function_map[f]["group"]
                    self.write_group_start(current_group)
                    self.write_empty_line()

                self.write_function(function_map[f])
                self.write_empty_line()

            # Write new functions
            self.write_empty_line(n=1)
            self.write_group_start("New Functions")
            self.write_empty_line(n=2)
            current_group = None
            for group in function_groups:
                for fn in group["entries"]:
                    if fn["name"] in self.manual_order:
                        continue
                    if current_group != group["group"]:
                        current_group = group["group"]
                        self.write_group_start(current_group)
                        self.write_empty_line()

                    self.write_function(fn)
                    self.write_empty_line()

        else:
            for group in function_groups:
                self.write_group_start(group["group"])
                self.write_empty_line()
                for fn in group["entries"]:
                    self.write_function(fn)
                    self.write_empty_line()
                self.write_empty_line()
                self.write_empty_line()

        self.write_footer()

    def _analyze_types(self, groups: List[FunctionGroup]):
        for group in groups:
            for fn in group["entries"]:
                for param in fn["params"]:
                    if param["type"] not in self.type_maps:
                        self.type_maps[param["type"]] = self._get_casted_type(param["type"])
                if fn["return_type"] not in self.type_maps:
                    self.type_maps[fn["return_type"]] = self._get_casted_type(fn["return_type"])

        for k, v in self.type_maps.items():
            if v not in self.inverse_type_maps:
                self.inverse_type_maps[v] = []
            self.inverse_type_maps[v].append(k)
        return

    def write_group_start(self, group):
        group = group.replace("_", " ").strip()
        # make group title uppercase
        group = " ".join([x.capitalize() for x in group.split(" ")])
        self.file.write(f"# {'-' * 80}\n")
        self.file.write(f"# {group}\n")
        self.file.write(f"# {'-' * 80}\n")

    @staticmethod
    def get_function_order(filepath):
        path = pathlib.Path(filepath)
        if not path.exists() or not path.is_file():
            raise FileNotFoundError(f"File {path} does not exist")

        with open(path, "r") as f:
            lines = f.readlines()

        is_julia_file = path.suffix == ".jl"

        if not is_julia_file:
            # read the file and assume that we have a function name per line
            return [x.strip() for x in lines if x.strip() != ""]

        # find the function definitions
        # TODO this a very simple regex that only supports the long function form `function name(...)`
        function_regex = r"^function\s+([a-zA-Z_][a-zA-Z0-9_]*)\s*\("
        function_order = []
        for line in lines:
            line = line.strip()
            if line.startswith("#"):
                continue

            m = re.match(function_regex, line)
            if m is not None:
                function_order.append(m.group(1))
        return function_order


def main():
    """Main function to generate the Julia API."""

    print("Creating Julia API")

    parser = configure_parser()
    args = parser.parse_args()
    print("Arguments:")
    for k, v in vars(args).items():
        print(f"    {k}: {v}")

    julia_path = pathlib.Path(args.output)
    enable_auto_1base_index = args.auto_1_index
    enable_original_order = args.use_original_order

    capi_defintions_dir = pathlib.Path(args.capi_dir)
    ext_api_definition_pattern = str(capi_defintions_dir) + "/apis/v1/*/*.json"
    capi_function_definition_pattern = str(capi_defintions_dir) + "/functions/**/*.json"
    ext_api_definitions = parse_ext_api_definitions(ext_api_definition_pattern)
    ext_api_version = get_extension_api_version(ext_api_definitions)
    function_groups, function_map = parse_capi_function_definitions(capi_function_definition_pattern)

    overwrite_function_signatures = {
        # Must be Ptr{Cvoid} and not Ref
        "duckdb_free": (
            "Cvoid",
            ("Ptr{Cvoid}",),
        ),
        "duckdb_bind_blob": (
            "duckdb_state",
            ("duckdb_prepared_statement", "idx_t", "Ptr{Cvoid}", "idx_t"),
        ),
        "duckdb_vector_assign_string_element_len": (
            "Cvoid",
            (
                "duckdb_vector",
                "idx_t",
                "Ptr{UInt8}",
                "idx_t",
            ),  # Must be Ptr{UInt8} instead of Cstring to allow '\0' in the middle
        ),
    }

    with JuliaApiTarget(
        julia_path,
        indent=0,
        auto_1base_index=enable_auto_1base_index,  # WARNING: every arg named "col/row/index" or similar will be 1-based indexed, so the argument is subtracted by 1
        auto_1base_index_return_functions={"duckdb_init_get_column_index"},
        auto_1base_index_ignore_functions={
            "duckdb_parameter_name",  # Parameter names start at 1
            "duckdb_param_type",  # Parameter types (like names) start at 1
            "duckdb_param_logical_type",  # ...
            "duckdb_bind_get_parameter",  # Would be breaking API change
        },
        skipped_functions={},
        type_map=JULIA_BASE_TYPE_MAP,
        overwrite_function_signatures=overwrite_function_signatures,
    ) as printer:
        if enable_original_order:
            print("INFO: Using the original order of the functions from the old API file.")
            printer.manual_order = JULIA_API_ORIGINAL_ORDER

        printer.write_functions(ext_api_version, function_groups, function_map)

        if args.print_type_mapping:
            print("Type maps: (Julia Type -> C Type)")
            K = list(printer.inverse_type_maps.keys())
            K.sort()
            for k in K:
                if k.startswith("Ptr") or k.startswith("Ref"):
                    continue
                v = ", ".join(printer.inverse_type_maps[k])
                print(f"    {k} -> {v}")

        print("Julia API generated successfully!")
        print("Please review the mapped types and check the generated file:")
        print("Hint: also run './format.sh' to format the file and reduce the diff.")
        print(f"Output: {julia_path}")


def configure_parser():
    parser = argparse.ArgumentParser(description="Generate the DuckDB Julia API")
    parser.add_argument(
        "--auto-1-index",
        action="store_true",
        default=True,
        help="Automatically convert 0-based indices to 1-based indices",
    )
    parser.add_argument(
        "--use-original-order",
        action="store_true",
        default=False,
        help="Use the original order of the functions from the old API file. New functions will be appended at the end.",
    )

    parser.add_argument(
        "--print-type-mapping",
        action="store_true",
        default=False,
        help="Print the type mapping from C to Julia",
    )

    parser.add_argument(
        "--capi-dir",
        type=str,
        required=True,
        help="Path to the input C API definitions. Should be a directory containing JSON files.",
    )
    parser.add_argument(
        "output",
        type=str,
        # default="src/api.jl",
        help="Path to the output file",
    )
    return parser


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Script to clean up JSON configuration files by removing skip_tests entries
that are already inherited from parent configurations.
"""

import json
import os
import sys
import argparse
from pathlib import Path


def load_config(filepath):
    """Load a JSON configuration file."""
    with open(filepath, 'r') as f:
        return json.load(f)


def save_config(filepath, config):
    """Save a JSON configuration file with proper formatting."""
    with open(filepath, 'w') as f:
        json.dump(config, f, indent=2)
        f.write('\n')


def get_inherited_skip_tests(config_path, base_dir, visited=None):
    """
    Recursively collect all inherited skip test paths.
    Returns a set of test paths.
    """
    if visited is None:
        visited = set()

    # Prevent infinite loops from circular inheritance
    config_path = os.path.abspath(config_path)
    if config_path in visited:
        return set()
    visited.add(config_path)

    config = load_config(config_path)
    inherited_paths = set()

    # Check if this config inherits from another
    if 'inherit_skip_tests' in config:
        inherit_path = config['inherit_skip_tests']
        # Resolve relative path from base directory (repo root)
        full_inherit_path = os.path.join(base_dir, inherit_path)

        if os.path.exists(full_inherit_path):
            # First, recursively get any tests inherited by the parent
            inherited_paths.update(get_inherited_skip_tests(full_inherit_path, base_dir, visited))

            # Then add the parent's own skip tests
            parent_config = load_config(full_inherit_path)
            if 'skip_tests' in parent_config:
                for skip_group in parent_config['skip_tests']:
                    if 'paths' in skip_group:
                        inherited_paths.update(skip_group['paths'])
        else:
            print(f"Warning: Inherited config not found: {full_inherit_path}", file=sys.stderr)

    return inherited_paths


def cleanup_config(config_path, base_dir, dry_run=False):
    """
    Clean up a config file by removing skip_tests that are already inherited.
    Returns True if changes were made, False otherwise.
    """
    config = load_config(config_path)

    # If no skip_tests or no inheritance, nothing to clean
    if 'skip_tests' not in config or 'inherit_skip_tests' not in config:
        return False

    inherited_tests = get_inherited_skip_tests(config_path, base_dir)

    if not inherited_tests:
        return False

    changes_made = False
    removed_tests = []

    # Process each skip_tests group
    new_skip_tests = []
    for skip_group in config['skip_tests']:
        if 'paths' not in skip_group:
            new_skip_tests.append(skip_group)
            continue

        # Filter out paths that are inherited
        original_paths = skip_group['paths']
        filtered_paths = [p for p in original_paths if p not in inherited_tests]

        # Track what was removed
        removed_from_group = [p for p in original_paths if p in inherited_tests]
        if removed_from_group:
            removed_tests.extend(removed_from_group)
            changes_made = True

        # Only keep the group if it still has paths
        if filtered_paths:
            new_group = skip_group.copy()
            new_group['paths'] = filtered_paths
            new_skip_tests.append(new_group)

    if changes_made:
        print(f"\n{config_path}:")
        print(f"  Removed {len(removed_tests)} inherited test(s):")
        for test in removed_tests:
            print(f"    - {test}")

        if not dry_run:
            config['skip_tests'] = new_skip_tests
            # Remove skip_tests entirely if empty
            if not new_skip_tests:
                del config['skip_tests']
            save_config(config_path, config)
            print("  Changes saved.")
        else:
            print("  (dry run - no changes saved)")

    return changes_made


def find_config_files(directory):
    """Find all JSON config files in the given directory."""
    config_files = []
    for filepath in Path(directory).rglob('*.json'):
        config_files.append(str(filepath))
    return config_files


def main():
    parser = argparse.ArgumentParser(description='Clean up JSON config files by removing inherited skip_tests.')
    parser.add_argument(
        'configs',
        nargs='*',
        help='Config files to clean up. If none specified, processes all JSON files in test/configs/',
    )
    parser.add_argument(
        '--base-dir',
        default=None,
        help='Base directory for resolving inherit_skip_tests paths (default: current directory)',
    )
    parser.add_argument(
        '--dry-run', '-n', action='store_true', help='Show what would be removed without making changes'
    )
    parser.add_argument('--verbose', '-v', action='store_true', help='Show more details')

    args = parser.parse_args()

    # Determine base directory
    base_dir = args.base_dir or os.getcwd()

    # Find config files to process
    if args.configs:
        config_files = args.configs
    else:
        # Default to test/configs directory
        config_dir = os.path.join(base_dir, 'test', 'configs')
        if os.path.exists(config_dir):
            config_files = find_config_files(config_dir)
        else:
            print(f"Error: Config directory not found: {config_dir}", file=sys.stderr)
            sys.exit(1)

    if not config_files:
        print("No config files found to process.")
        return

    if args.verbose:
        print(f"Processing {len(config_files)} config file(s)...")
        if args.dry_run:
            print("(dry run mode)")

    total_changed = 0
    for config_path in sorted(config_files):
        try:
            if cleanup_config(config_path, base_dir, args.dry_run):
                total_changed += 1
        except json.JSONDecodeError as e:
            print(f"Error parsing {config_path}: {e}", file=sys.stderr)
        except Exception as e:
            print(f"Error processing {config_path}: {e}", file=sys.stderr)

    if total_changed == 0:
        print("\nNo redundant inherited tests found.")
    else:
        print(f"\n{'Would clean' if args.dry_run else 'Cleaned'} {total_changed} file(s).")


if __name__ == '__main__':
    main()



# fmt: off

import pytest
import subprocess
import sys
from typing import List
from conftest import ShellTest


def test_profiling_json(shell, tmp_path):
    target_dir = tmp_path / 'export_test'
    test = (
        ShellTest(shell)
        .statement(".mode csv")
        .statement(".changes off")
        .statement("CREATE TABLE integers(i INTEGER);")
        .statement("CREATE TABLE integers2(i INTEGER);")
        .statement("INSERT INTO integers SELECT * FROM range(100);")
        .statement("INSERT INTO integers2 VALUES (1), (3), (99);")
        .statement(f"EXPORT DATABASE '{target_dir.as_posix()}';")
        .statement("DROP TABLE integers;")
        .statement("DROP TABLE integers2;")
        .statement(f"IMPORT DATABASE '{target_dir.as_posix()}';")
        .statement("SELECT SUM(i)*MAX(i) FROM integers JOIN integers2 USING (i);")
    )
    result = test.run()
    result.check_stdout('10197')

# fmt: on



# use bison to generate the parser files
# the following version of bison is used:
# bison (GNU Bison) 2.3
import os
import subprocess
import re
import sys
from python_helpers import open_utf8

bison_location = "bison"
base_dir = 'third_party/libpg_query/grammar'
pg_dir = 'third_party/libpg_query'
namespace = 'duckdb_libpgquery'

counterexamples = False
run_update = False
verbose = False
for arg in sys.argv[1:]:
    if arg.startswith("--bison="):
        bison_location = arg.replace("--bison=", "")
    elif arg.startswith("--counterexamples"):
        counterexamples = True
    elif arg.startswith("--update"):
        run_update = True
    # allow a prefix to the source and target directories
    elif arg.startswith("--custom_dir_prefix"):
        base_dir = arg.split("=")[1] + base_dir
        pg_dir = arg.split("=")[1] + pg_dir
    elif arg.startswith("--namespace"):
        namespace = arg.split("=")[1]
    elif arg.startswith("--verbose"):
        verbose = True
    else:
        raise Exception(
            "Unrecognized argument: "
            + arg
            + ", expected --counterexamples, --bison=/loc/to/bison, --custom_dir_prefix, --namespace, --verbose"
        )

template_file = os.path.join(base_dir, 'grammar.y')
target_file = os.path.join(base_dir, 'grammar.y.tmp')
header_file = os.path.join(base_dir, 'grammar.hpp')
source_file = os.path.join(base_dir, 'grammar.cpp')
type_dir = os.path.join(base_dir, 'types')
rule_dir = os.path.join(base_dir, 'statements')
result_source = os.path.join(base_dir, 'grammar_out.cpp')
result_header = os.path.join(base_dir, 'grammar_out.hpp')
target_source_loc = os.path.join(pg_dir, 'src_backend_parser_gram.cpp')
target_header_loc = os.path.join(pg_dir, 'include/parser/gram.hpp')
kwlist_header = os.path.join(pg_dir, 'include/parser/kwlist.hpp')


# parse the keyword lists
def read_list_from_file(fname):
    with open_utf8(fname, 'r') as f:
        return [x.strip() for x in f.read().split('\n') if len(x.strip()) > 0]


kwdir = os.path.join(base_dir, 'keywords')
unreserved_keywords = read_list_from_file(os.path.join(kwdir, 'unreserved_keywords.list'))
colname_keywords = read_list_from_file(os.path.join(kwdir, 'column_name_keywords.list'))
func_name_keywords = read_list_from_file(os.path.join(kwdir, 'func_name_keywords.list'))
type_name_keywords = read_list_from_file(os.path.join(kwdir, 'type_name_keywords.list'))
reserved_keywords = read_list_from_file(os.path.join(kwdir, 'reserved_keywords.list'))


def strip_p(x):
    if x.endswith("_P"):
        return x[:-2]
    else:
        return x


unreserved_keywords.sort(key=lambda x: strip_p(x))
colname_keywords.sort(key=lambda x: strip_p(x))
func_name_keywords.sort(key=lambda x: strip_p(x))
type_name_keywords.sort(key=lambda x: strip_p(x))
reserved_keywords.sort(key=lambda x: strip_p(x))

statements = read_list_from_file(os.path.join(base_dir, 'statements.list'))
statements.sort()
if len(statements) < 0:
    print("Need at least one statement")
    exit(1)

# verify there are no duplicate keywords and create big sorted list of keywords
kwdict = {}
for kw in unreserved_keywords:
    kwdict[kw] = 'UNRESERVED_KEYWORD'

for kw in colname_keywords:
    kwdict[kw] = 'COL_NAME_KEYWORD'

for kw in func_name_keywords:
    kwdict[kw] = 'TYPE_FUNC_NAME_KEYWORD'

for kw in type_name_keywords:
    kwdict[kw] = 'TYPE_FUNC_NAME_KEYWORD'

for kw in reserved_keywords:
    kwdict[kw] = 'RESERVED_KEYWORD'

kwlist = [(x, kwdict[x]) for x in kwdict.keys()]
# sorting uppercase is different from lowercase: A-Z < _ < a-z
kwlist.sort(key=lambda x: strip_p(x[0].lower()))

# now generate kwlist.h
# PG_KEYWORD("abort", ABORT_P, UNRESERVED_KEYWORD)
kwtext = (
    """
namespace """
    + namespace
    + """ {
#define PG_KEYWORD(a,b,c) {a,b,c},

const PGScanKeyword ScanKeywords[] = {
"""
)
for tpl in kwlist:
    kwtext += 'PG_KEYWORD("%s", %s, %s)\n' % (strip_p(tpl[0]).lower(), tpl[0], tpl[1])
kwtext += (
    """
};

const int NumScanKeywords = lengthof(ScanKeywords);
} // namespace """
    + namespace
    + """
"""
)

with open_utf8(kwlist_header, 'w+') as f:
    f.write(kwtext)


# generate the final main.y.tmp file
# first read the template file
with open_utf8(template_file, 'r') as f:
    text = f.read()

# now perform a series of replacements in the file to construct the final yacc file


def get_file_contents(fpath, add_line_numbers=False):
    with open_utf8(fpath, 'r') as f:
        result = f.read()
        if add_line_numbers:
            return '#line 1 "%s"\n' % (fpath,) + result
        else:
            return result


# grammar.hpp
text = text.replace("{{{ GRAMMAR_HEADER }}}", get_file_contents(header_file, True))

# grammar.cpp
text = text.replace("{{{ GRAMMAR_SOURCE }}}", get_file_contents(source_file, True))

# keyword list
kw_token_list = "%token <keyword> " + " ".join([x[0] for x in kwlist])

text = text.replace("{{{ KEYWORDS }}}", kw_token_list)

# statements
stmt_list = "stmt: " + "\n\t| ".join(statements) + "\n\t| /*EMPTY*/\n\t{ $$ = NULL; }\n"
text = text.replace("{{{ STATEMENTS }}}", stmt_list)

# keywords
# keywords can EITHER be reserved, unreserved, or some combination of (col_name, type_name, func_name)
# that means duplicates are ONLY allowed between (col_name, type_name and func_name)
# having a keyword be both reserved and unreserved is an error
# as is having a keyword both reserved and col_name, for example
# verify that this is the case
reserved_dict = {}
unreserved_dict = {}
other_dict = {}
for r in reserved_keywords:
    if r in reserved_dict:
        print("Duplicate keyword " + r + " in reserved keywords")
        exit(1)
    reserved_dict[r] = True

for ur in unreserved_keywords:
    if ur in unreserved_dict:
        print("Duplicate keyword " + ur + " in unreserved keywords")
        exit(1)
    if ur in reserved_dict:
        print("Keyword " + ur + " is marked as both unreserved and reserved")
        exit(1)
    unreserved_dict[ur] = True


def add_to_other_keywords(kw, list_name):
    global unreserved_dict
    global reserved_dict
    global other_dict
    if kw in unreserved_dict:
        print("Keyword " + kw + " is marked as both unreserved and " + list_name)
        exit(1)
    if kw in reserved_dict:
        print("Keyword " + kw + " is marked as both reserved and " + list_name)
        exit(1)
    other_dict[kw] = True


for cr in colname_keywords:
    add_to_other_keywords(cr, "colname")

type_func_name_dict = {}
for tr in type_name_keywords:
    add_to_other_keywords(tr, "typename")
    type_func_name_dict[tr] = True

for fr in func_name_keywords:
    add_to_other_keywords(fr, "funcname")
    type_func_name_dict[fr] = True

type_func_name_keywords = list(type_func_name_dict.keys())
type_func_name_keywords.sort()

all_keywords = list(reserved_dict.keys()) + list(unreserved_dict.keys()) + list(other_dict.keys())
all_keywords.sort()

other_keyword = list(other_dict.keys())
other_keyword.sort()

kw_definitions = "unreserved_keyword: " + " | ".join(unreserved_keywords) + "\n"
kw_definitions += "col_name_keyword: " + " | ".join(colname_keywords) + "\n"
kw_definitions += "func_name_keyword: " + " | ".join(func_name_keywords) + "\n"
kw_definitions += "type_name_keyword: " + " | ".join(type_name_keywords) + "\n"
kw_definitions += "other_keyword: " + " | ".join(other_keyword) + "\n"
kw_definitions += "type_func_name_keyword: " + " | ".join(type_func_name_keywords) + "\n"
kw_definitions += "reserved_keyword: " + " | ".join(reserved_keywords) + "\n"
text = text.replace("{{{ KEYWORD_DEFINITIONS }}}", kw_definitions)


# types
def concat_dir(dname, extension, add_line_numbers=False):
    result = ""
    for fname in os.listdir(dname):
        fpath = os.path.join(dname, fname)
        if os.path.isdir(fpath):
            result += concat_dir(fpath, extension)
        else:
            if not fname.endswith(extension):
                continue
            result += get_file_contents(fpath, add_line_numbers)
    return result


type_definitions = concat_dir(type_dir, ".yh")
# add statement types as well
for stmt in statements:
    type_definitions += "%type <node> " + stmt + "\n"

text = text.replace("{{{ TYPES }}}", type_definitions)

# grammar rules
grammar_rules = concat_dir(rule_dir, ".y", True)

text = text.replace("{{{ GRAMMAR RULES }}}", grammar_rules)

# finally write the yacc file into the target file
with open_utf8(target_file, 'w+') as f:
    f.write(text)

# generate the bison
cmd = [bison_location]
if counterexamples:
    print("Attempting to print counterexamples (-Wcounterexamples)")
    cmd += ["-Wcounterexamples"]
if run_update:
    cmd += ["--update"]
if verbose:
    cmd += ["--verbose"]
cmd += ["-o", result_source, "-d", target_file]
print(' '.join(cmd))
proc = subprocess.Popen(cmd, stderr=subprocess.PIPE)
res = proc.wait(timeout=10)  # ensure CI does not hang as was seen when running with Bison 3.x release.

if res != 0:
    text = proc.stderr.read().decode('utf8')
    print(text)
    if 'shift/reduce' in text and not counterexamples:
        print("---------------------------------------------------------------------")
        print("In case of shift/reduce conflicts, try re-running with --counterexamples")
        print("Note: this requires a more recent version of Bison (e.g. version 3.8)")
        print("On a Macbook you can obtain this using \"brew install bison\"")
    if counterexamples and 'time limit exceeded' in text:
        print("---------------------------------------------------------------------")
        print(
            "The counterexamples time limit was exceeded. This likely means that no useful counterexample was generated."
        )
        print("")
        print("The counterexamples time limit can be increased by setting the TIME_LIMIT environment variable, e.g.:")
        print("export TIME_LIMIT=100")
    exit(1)


os.rename(result_source, target_source_loc)
os.rename(result_header, target_header_loc)

with open_utf8(target_source_loc, 'r') as f:
    text = f.read()

text = text.replace('#include "grammar_out.hpp"', '#include "include/parser/gram.hpp"')
text = text.replace('yynerrs = 0;', 'yynerrs = 0; (void)yynerrs;')

with open_utf8(target_source_loc, 'w+') as f:
    f.write(text)
